/***********************************************************************
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 ***********************************************************************/
#include <linux/init.h>
#include <linux/cpufreq.h>
#include <linux/spinlock.h>

#include <asm/irq.h>
#include <asm/io.h>
#include <asm/mach/irq.h>

#include <asm/arch/ns9750_sys.h>
#include <asm/arch/ns9750_bbus.h>
#include <asm/arch/ns9750_bbus_bridge.h>
#include <asm/arch/ns9750_usb.h>

#if defined(CONFIG_MACH_ARMADILLO300)
#define MISCREG_VIRT_BASE       (0xf6000000)
#define MISCREG_BASE            MISCREG_VIRT_BASE
#define MISCREG(type,off) (*(volatile type *)(MISCREG_BASE + (off)))
#define MISCREG_INTCR     MISCREG(u8, 0x0)
#define MISCREG_INTMR     MISCREG(u8, 0x2)
volatile u8 miscreg_intmr  = 0x0;
#endif

static struct resource resources[] = {
	{
		.name	= "irq_ahb",
		.start	= (NS_SYS_MODULE_BASE + NS_SYS_INT_VEC_ADR_BASE),
		.end	= (NS_SYS_MODULE_BASE + NS_SYS_INT_STAT_RAW + 3)
	}, {
		.name	= "irq_bbus_bridge",
		.start	= (NS_BBUS_BRIDGE_MODULE_BASE + NS_BBUS_BRIDGE_ISTAT),
		.end	= (NS_BBUS_BRIDGE_MODULE_BASE + NS_BBUS_BRIDGE_IEN + 3)
	}, {
		.name	= "irq_bbus",
		.start	= (NS_BBUS_MODULE_BASE + NS_BBUS_DMA_ISTAT),
		.end	= (NS_BBUS_MODULE_BASE + NS_BBUS_DMA_IEN + 3)
	}, {
		.name	= "irq_usb",
		.start	= (NS_USB_MODULE_BASE + NS_USB_IEN),
		.end	= (NS_USB_MODULE_BASE + NS_USB_ISTAT + 3)
	}, {
#if defined(CONFIG_MACH_ARMADILLO300)
		.name	= "irq_ext0",
		.start	= (0xf7000000),
		.end	= (0xf7000030)
	}, {
#endif
		.name = NULL
	}
};

static inline void ns_disable_ahb_irq( unsigned int irq );
static inline void ns_enable_ahb_irq( unsigned int irq );
static inline void ns_ack_ahb_irq( unsigned int irq );

static inline void ns_disable_bbus_irq( unsigned int irq );
static inline void ns_enable_bbus_irq( unsigned int irq );
static inline void ns_ack_bbus_irq( unsigned int irq );

static inline void ns_disable_bbus_dma_irq( unsigned int irq );
static inline void ns_enable_bbus_dma_irq( unsigned int irq );
static inline void ns_ack_bbus_dma_irq( unsigned int irq );

static inline void ns_disable_usb_irq( unsigned int irq );
static inline void ns_enable_usb_irq( unsigned int irq );
static inline void ns_ack_usb_irq( unsigned int irq );

#if defined(CONFIG_MACH_ARMADILLO300)
static inline void ns_disable_ext0_irq( unsigned int irq );
static inline void ns_enable_ext0_irq( unsigned int irq );
static inline void ns_ack_ext0_irq( unsigned int irq );

static void ns_ext0_irq_handler( unsigned int irq,
				 struct irqdesc *desc,
				 struct pt_regs *regs );
#endif

static struct irqchip ns_ahb_irq_chip = {
	.ack	= ns_ack_ahb_irq,
	.mask	= ns_disable_ahb_irq,
	.unmask = ns_enable_ahb_irq
};

static struct irqchip ns_bbus_irq_chip = {
	.ack	= ns_ack_bbus_irq,
	.mask	= ns_disable_bbus_irq,
	.unmask = ns_enable_bbus_irq
};

static struct irqchip ns_bbus_dma_irq_chip = {
	.ack	= ns_ack_bbus_dma_irq,
	.mask	= ns_disable_bbus_dma_irq,
	.unmask = ns_enable_bbus_dma_irq
};

static struct irqchip ns_usb_irq_chip = {
	.ack	= ns_ack_usb_irq,
	.mask	= ns_disable_usb_irq,
	.unmask = ns_enable_usb_irq
};

#if defined(CONFIG_MACH_ARMADILLO300)
static struct irqchip ns_ext0_irq_chip = {
	.ack	= ns_ack_ext0_irq,
	.mask	= ns_disable_ext0_irq,
	.unmask	= ns_enable_ext0_irq
};
#endif

spinlock_t ns_irq_lock;

void __init ns_init_irq( void )
{
	int i = 0;

	while( resources[ i ].name != NULL ) {
		request_resource( &iomem_resource, &resources[ i++ ] );
	}

	spin_lock_init( &ns_irq_lock );

	/* disable all IRQs */
	for( i = 0; i < 8; i++ )
		writel( 0, NS_SYS_INT_CFG_BASE + i*4 );

	for( i = 0; i < 32; i++ )
		writel( i, NS_SYS_INT_VEC_ADR( i ) );

	/* configure linux interrupt handling */
	for( i = 0; i < NR_IRQ_BBUS_START; i++ ) {
		set_irq_chip( i, &ns_ahb_irq_chip );
		set_irq_handler( i, do_level_IRQ );
		set_irq_flags( i, IRQF_VALID | IRQF_PROBE );
	}
	for( i = NR_IRQ_BBUS_START; i < NR_IRQ_BBUS_DMA_START; i++ ) {
		set_irq_chip( i, &ns_bbus_irq_chip );
		set_irq_handler( i, do_level_IRQ );
		set_irq_flags( i, IRQF_VALID | IRQF_PROBE );
	}

	for( i = NR_IRQ_BBUS_DMA_START; i < NR_IRQ_USB_START; i++ ) {
		set_irq_chip( i, &ns_bbus_dma_irq_chip );
		set_irq_handler( i, do_level_IRQ );
		set_irq_flags( i, IRQF_VALID | IRQF_PROBE );
	}

	for( i = NR_IRQ_USB_START; i < NR_IRQ_EXTINT0_START; i++ ) {
		set_irq_chip( i, &ns_usb_irq_chip );
		set_irq_handler( i, do_level_IRQ );
		set_irq_flags( i, IRQF_VALID | IRQF_PROBE );
	}

	for( i = NR_IRQ_EXTINT0_START; i < NR_IRQS; i++ ) {
		set_irq_chip( i, &ns_ext0_irq_chip );
		set_irq_handler( i, do_level_IRQ );
		set_irq_flags( i, IRQF_VALID | IRQF_PROBE );
	}
	set_irq_chained_handler( IRQ_EXT_0, ns_ext0_irq_handler );

	/* disable all BBUS Interrupts all but DMA controller but propagate
	 * later enabled ones to sys module */
	writel( NS_BBUS_BRIDGE_IEN_GLBL |
		NS_BBUS_BRIDGE_IEN_USB  |
		NS_BBUS_BRIDGE_IEN_BBDMA,
		NS_BBUS_BRIDGE_IEN );
	/* we need no handler for it. IRQ_BBUS_AGGREGATE will trigger the BBUS
	 * handlers immediately in entry-armv.S */
	ns_enable_ahb_irq( IRQ_BBUS_AGGREGATE );
	ns_enable_bbus_irq( IRQ_BBUS_DMA_AGG );

#if defined(CONFIG_MACH_ARMADILLO300)
	MISCREG_INTMR = EXTINT0_IRQS;
	MISCREG_INTCR = EXTINT0_IRQS;
#endif
}

static inline void ns_disable_ahb_irq( unsigned int irq )
{
	unsigned int reg;
	unsigned int byte;
	unsigned long flags;

	reg = irq / 4;
	byte = 3-(irq % 4);	/* INT_CFG is big endian, buggy chip! */

	spin_lock_irqsave( &ns_irq_lock, flags );
	rmwl( NS_SYS_INT_CFG_BASE + reg*4,
	      & ~(NS_SYS_INT_CFG_IE << (byte*8) ) );
	spin_unlock_irqrestore( &ns_irq_lock, flags );
}

static inline void ns_enable_ahb_irq( unsigned int irq )
{
	unsigned int reg;
	unsigned int byte;
	unsigned int int_cfg;
	unsigned long flags;

	reg = irq / 4;
	byte = 3-(irq % 4);	/* INT_CFG is big endian, buggy chip! */

	/* @TODO: maybe we should set up some priority (timer first, then
	 * ethernet? */
	int_cfg = NS_SYS_INT_CFG_IE | irq;

	if(irq == IRQ_ETH_PHY){
		int_cfg |= NS_SYS_INT_CFG_INV;
	}

	spin_lock_irqsave( &ns_irq_lock, flags );
	rmwl( NS_SYS_INT_CFG_BASE + reg*4, |int_cfg << (byte*8) );
	spin_unlock_irqrestore( &ns_irq_lock, flags );
}


static inline void ns_ack_ahb_irq( /*@unused@*/ unsigned int irq )
{
 	ns_disable_ahb_irq( irq );
	writel( 0, NS_SYS_ISRADDR ); /* [1] p. 125 */
}

static inline void ns_disable_bbus_irq( unsigned int irq )
{
	unsigned long flags;

	spin_lock_irqsave( &ns_irq_lock, flags );
	rmwl( NS_BBUS_BRIDGE_IEN,
	      & ~(1 << (irq - NR_IRQ_BBUS_START)));
	spin_unlock_irqrestore( &ns_irq_lock, flags );
}

static inline void ns_enable_bbus_irq( unsigned int irq )
{
	unsigned long flags;

	spin_lock_irqsave( &ns_irq_lock, flags );
	/* enable BBUS IRQ */
	rmwl( NS_BBUS_BRIDGE_IEN,
	      | 1 << (irq - NR_IRQ_BBUS_START) );
	spin_unlock_irqrestore( &ns_irq_lock, flags );
	ns_enable_ahb_irq( IRQ_BBUS_AGGREGATE );  
}

static inline void ns_ack_bbus_irq( /*@unused@*/ unsigned int irq )
{
	ns_disable_bbus_irq( irq );
	ns_disable_ahb_irq( IRQ_BBUS_AGGREGATE );
	writel( 0, NS_SYS_ISRADDR ); /* [1] p. 125 */
}

static inline void ns_disable_bbus_dma_irq( unsigned int irq )
{
	unsigned long flags;

	spin_lock_irqsave( &ns_irq_lock, flags );
	rmwl( NS_BBUS_DMA_IEN,
	      & ~(1 << (irq - NR_IRQ_BBUS_DMA_START) ) );
	spin_unlock_irqrestore( &ns_irq_lock, flags );
}

static inline void ns_enable_bbus_dma_irq( unsigned int irq )
{
	unsigned long flags;

	spin_lock_irqsave( &ns_irq_lock, flags );
	rmwl( NS_BBUS_DMA_IEN,
	      | 1 << (irq - NR_IRQ_BBUS_DMA_START) );
	spin_unlock_irqrestore( &ns_irq_lock, flags );
}

static inline void ns_ack_bbus_dma_irq( /*@unused@*/ unsigned int irq )
{
	ns_disable_bbus_dma_irq( irq );
	writel( 0, NS_SYS_ISRADDR ); /* [1] p. 125 */
}

static inline void ns_disable_usb_irq( unsigned int irq )
{
	unsigned long flags;
	unsigned int mask = 0;

	spin_lock_irqsave( &ns_irq_lock, flags );
	/* don't allow selection of individual DMA interrupts. */
	if( irq == IRQ_USB_DMA )
		mask = NS_USB_INT_DMA_ALL;
	else if( irq == IRQ_USB_FIFO )
		mask = NS_USB_INT_FIFO_ALL;
	else if( irq == IRQ_USB_HOST )
		mask = NS_USB_INT_HOST;

	writel( (readl( NS_USB_IEN ) & ~mask) |	NS_USB_INT_GBL_EN,
		NS_USB_IEN );
	spin_unlock_irqrestore( &ns_irq_lock, flags );
}

static inline void ns_enable_usb_irq( unsigned int irq )
{
	unsigned long flags;
	unsigned int mask = 0;

	spin_lock_irqsave( &ns_irq_lock, flags );
	/* don't allow selection of individual DMA interrupts. */
	if( irq == IRQ_USB_DMA )
		mask = NS_USB_INT_DMA_ALL;
	else if( irq == IRQ_USB_FIFO )
		mask = NS_USB_INT_FIFO_ALL;
	else if( irq == IRQ_USB_HOST )
		mask = NS_USB_INT_HOST;

	rmwl( NS_USB_IEN, | (mask | NS_USB_INT_GBL_EN) );
	spin_unlock_irqrestore( &ns_irq_lock, flags );
}

static inline void ns_ack_usb_irq( /*@unused@*/ unsigned int irq )
{
	ns_disable_usb_irq( irq );
	writel( 0, NS_SYS_ISRADDR ); /* [1] p. 125 */
}

#if defined(CONFIG_MACH_ARMADILLO300)
static inline void ns_disable_ext0_irq( unsigned int irq )
{
        u8 mask;

        mask = 1 << (irq - NR_IRQ_EXTINT0_START);
        miscreg_intmr |= mask;
        MISCREG_INTMR = miscreg_intmr;
}

static inline void ns_enable_ext0_irq( unsigned int irq )
{
        miscreg_intmr &= ~(1 << (irq - NR_IRQ_EXTINT0_START));
        MISCREG_INTMR = miscreg_intmr;
}

static inline void ns_ack_ext0_irq( unsigned int irq )
{
        u8 mask;

	readl(NS_SYS_ISRADDR);
        mask = 1 << (irq - NR_IRQ_EXTINT0_START);
        miscreg_intmr |= mask;
        MISCREG_INTMR = miscreg_intmr;
        MISCREG_INTCR = mask;
	writel(0, NS_SYS_ISRADDR);
}

static void ns_ext0_irq_handler( unsigned int irq,
				 struct irqdesc *desc,
				 struct pt_regs *regs )
{
  unsigned char status;
  int i;

  status = MISCREG_INTCR & EXTINT0_IRQS;

  for(i=0;i<8 && status != 0;i++){
    int irqno;

    if(status & 0x01) {
      irqno = NR_IRQ_EXTINT0_START + i;
      desc->handle(irqno, &irq_desc[irqno], regs);
    }
    status  >>= 1;
  }
}
#endif
