/***********************************************************************
 * drivers/watchdog/ns9750_wdt.c
 *
 * Copyright (C) 2005 by FS Forth-Systeme GmbH.
 * All rights reserved.
 *
 * $Id: ns9750_wdt.c,v 1.4 2005/09/19 07:53:48 jdietsch Exp $
 * @Author: Jonas Dietsche
 * @Descr: watchdog driver
 * @References: [1]
 *			    [2] ixp4xx_wdt.c
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 ***********************************************************************/
#include <linux/config.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/miscdevice.h>
#include <linux/watchdog.h>
#include <linux/init.h>
#include <linux/bitops.h>

#include <asm/arch/ns9750_sys.h>
#ifdef CONFIG_ARCH_NS9750
#include <asm/arch/ns9750_config.h>
#elif CONFIG_ARCH_NS9360
#include <asm/arch/ns9360_config.h>
#endif
#include <asm/arch/irqs.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <linux/interrupt.h>
#include <asm/uaccess.h>

#ifdef CONFIG_WATCHDOG_NOWAYOUT
static int nowayout = 1;
#else
static int nowayout = 0;
#endif
static unsigned int heartbeat = 60;	/* (secs) Default is 1 minute */
static unsigned long wdt_status;

#define	WDT_IN_USE		0
#define	WDT_OK_TO_CLOSE		1
#define 	MAX_TIMER	0xffffffff
#define	DRIVER_NAME		"NS9xx0 Watchdog"

static unsigned long wdt_tick_rate;

static void
wdt_enable(void)
{
	rmwl( NS_SYS_SW_WDOG_CFG, | NS_SYS_SW_WDOG_CFG_SWWE);
}

static void
wdt_disable(void)
{
	/* cannot be stopped */
}

static int
wdt_keepalive(void)
{
	int i=2;
	int j=0;

	while( heartbeat >= (MAX_TIMER/(NS_SYS_CLK_FREQ/2/i)) && i<=64)
	{
		i*=2;
		j++;
	}
	wdt_tick_rate = NS_SYS_CLK_FREQ/2/i;
	rmwl( NS_SYS_SW_WDOG_CFG, | (NS_SYS_SW_WDOG_CFG_SWTCS_2 +j));

	writel( heartbeat * wdt_tick_rate, NS_SYS_SW_WDOG_TIMER);

	return 0;
}

static int
ns9750_wdt_open(struct inode *inode, struct file *file)
{
	if (test_and_set_bit(WDT_IN_USE, &wdt_status))
		return -EBUSY;

	clear_bit(WDT_OK_TO_CLOSE, &wdt_status);

	wdt_enable();

	return nonseekable_open(inode, file);
}

static ssize_t
ns9750_wdt_write(struct file *file, const char *data, size_t len, loff_t *ppos)
{
	if (len) {
		if (!nowayout) {
			size_t i;

			clear_bit(WDT_OK_TO_CLOSE, &wdt_status);

			for (i = 0; i != len; i++) {
				char c;

				if (get_user(c, data + i))
					return -EFAULT;
				if (c == 'V')
					set_bit(WDT_OK_TO_CLOSE, &wdt_status);
			}
		}
		wdt_keepalive();
	}

	return len;
}


static struct watchdog_info ident = {
	.options	= WDIOF_MAGICCLOSE | WDIOF_SETTIMEOUT |
				WDIOF_KEEPALIVEPING,
	.identity	= "NS9xx0 Watchdog",
};

static int
ns9750_wdt_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
			unsigned long arg)
{
	int ret = -ENOIOCTLCMD;
	int time;

	switch (cmd) {
	case WDIOC_GETSUPPORT:
		ret = copy_to_user((struct watchdog_info *)arg, &ident,
				   sizeof(ident)) ? -EFAULT : 0;
		break;

	case WDIOC_GETSTATUS:
		ret = put_user(0, (int *)arg);
		break;

	case WDIOC_GETBOOTSTATUS:
		ret = put_user(0, (int *)arg);
		break;

	case WDIOC_SETTIMEOUT:
		ret = get_user(time, (int *)arg);
		if (ret)
			break;

		if (time <= 0 || time > 1374) {
			ret = -EINVAL;
			break;
		}

		heartbeat = time;
		wdt_keepalive();
		/* Fall through */

	case WDIOC_GETTIMEOUT:
		ret = put_user(heartbeat, (int *)arg);
		break;

	case WDIOC_KEEPALIVE:
		wdt_enable();
		ret = 0;
		break;
	}

	return ret;
}

static int
ns9750_wdt_release(struct inode *inode, struct file *file)
{
	if (test_bit(WDT_OK_TO_CLOSE, &wdt_status)) {
		wdt_disable();
	} else {
		printk(KERN_CRIT "%s: Device closed unexpectedly - "
					"timer will not stop\n", DRIVER_NAME);
	}

	clear_bit(WDT_IN_USE, &wdt_status);
	clear_bit(WDT_OK_TO_CLOSE, &wdt_status);

	return 0;
}

static irqreturn_t ns9750_wdt_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	/* still alive, reset timer */
	wdt_keepalive();

	/* we have to write a 1, then a 0 to SWWI to clear the interrupt */
	rmwl( NS_SYS_SW_WDOG_CFG, | NS_SYS_SW_WDOG_CFG_SWWI);
	rmwl( NS_SYS_SW_WDOG_CFG, & ~NS_SYS_SW_WDOG_CFG_SWWI);

	return IRQ_HANDLED;
}

static struct file_operations ns9750_wdt_fops =
{
	.owner		= THIS_MODULE,
	.llseek		= no_llseek,
	.write		= ns9750_wdt_write,
	.ioctl		= ns9750_wdt_ioctl,
	.open		= ns9750_wdt_open,
	.release	= ns9750_wdt_release,
};

static struct miscdevice ns9750_wdt_miscdev =
{
	.minor		= WATCHDOG_MINOR,
	.name		= "ns9xx0_wdt",
	.fops		= &ns9750_wdt_fops,
};

static int __init ns9750_wdt_init(void)
{
	wdt_tick_rate = NS_SYS_CLK_FREQ/2/4;
	request_irq (IRQ_WATCHDOG, ns9750_wdt_interrupt, SA_INTERRUPT,
			      DRIVER_NAME, 0);

#ifdef CONFIG_NS9750_WATCHDOG_RESET
	rmwl( NS_SYS_SW_WDOG_CFG, | NS_SYS_SW_WDOG_CFG_SWWIC);
#endif /* CONFIG_NS9750_WATCHDOG_RESET */

#ifdef CONFIG_NS9750_WATCHDOG_BOOT
	rmwl( NS_SYS_SW_WDOG_CFG, | NS_SYS_SW_WDOG_CFG_SWWE);
#endif /* CONFIG_NS9750_WATCHDOG_BOOT */

	printk( KERN_INFO "%s initialized\n", DRIVER_NAME );

	return misc_register(&ns9750_wdt_miscdev);
}

static void __exit ns9750_wdt_exit(void)
{
	free_irq( IRQ_WATCHDOG, 0 );
	misc_deregister(&ns9750_wdt_miscdev);
}

module_init(ns9750_wdt_init);
module_exit(ns9750_wdt_exit);

MODULE_AUTHOR("Jonas Dietsche">);
MODULE_DESCRIPTION("NS9xx0 Watchdog Driver");

module_param(heartbeat, int, 0);
MODULE_PARM_DESC(heartbeat, "Watchdog heartbeat in seconds (default 60s)");

module_param(nowayout, int, 0);
MODULE_PARM_DESC(nowayout, "Watchdog cannot be stopped once started");

MODULE_LICENSE("GPL");
MODULE_ALIAS_MISCDEV(WATCHDOG_MINOR);

