/*****************************************************************************
 *
 * ns9750_eth.c
 *    NS9750 Ethernet Driver
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 *****************************************************************************/
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/dma-mapping.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>

#include <asm/io.h>

#include <asm/arch/ns9750_eth.h>

#define DEVICE_NAME "ns9750ethernet"
#define DRIVER_NAME "ns9750_eth"

//#define DEBUG_ETH
#if defined(DEBUG_ETH)
#define DEBUG_FUNC()		printk(DRIVER_NAME ": %s()\n", __FUNCTION__)
#define DEBUG_WARN(args...)	printk(DRIVER_NAME ": " args)
#define DEBUG_INFO(args...)	printk(DRIVER_NAME ": " args)
#else
#define DEBUG_FUNC()
#define DEBUG_WARN(args...)
#define DEBUG_INFO(args...)
#endif
#define PRINT_ERR(args...)	printk(KERN_ERR DRIVER_NAME ": " args)
#define PRINT_WARN(args...)	printk(KERN_WARNING DRIVER_NAME ": " args)
#define PRINT_INFO(args...)	printk(KERN_INFO DRIVER_NAME ": " args)

#define PRIV(dev)		((struct ns9750_eth_info *)dev->priv)

#define RegRd8(ofs)		*(volatile u8  *)(ofs)
#define RegRd16(ofs)		*(volatile u16 *)(ofs)
#define RegRd32(ofs)		*(volatile u32 *)(ofs)
#define RegWr8(ofs, val)	*(volatile u8  *)(ofs)=(u8)(val)
#define RegWr16(ofs, val)	*(volatile u16 *)(ofs)=(u16)(val)
#define RegWr32(ofs, val)	*(volatile u32 *)(ofs)=(u32)(val)

#define NS_ETH_PHY_ADDRESS	(0x0000)

#define NS_RX_RINGS		(4)
#define NS_RX_DESC_RING 	(64)
#define NS_RX_DESC		(NS_RX_RINGS * NS_RX_DESC_RING)
#define NS_TX_DESC		(64)

#define NS_ETH_FRAME_LEN	(0x600)

#define NS_MII_POLE_TIMEOUT	(10000)
#define NS_MII_NEG_DELAY	(5*HZ)

#define NS_POLYNOMIAL		(0x4c11db6L) /* taken from p407 */
#define NS_LEFT			(0)
#define NS_RIGHT		(1)

#define LINK_STATUS_UNKNOWN	(0xffff)

typedef struct rx_buffer_desc_t{
	unsigned int *src;
	unsigned int len;
	unsigned int *dest;
	union{
		unsigned int reg;
		struct{
			unsigned status : 16;
			unsigned res    : 12;
			unsigned full   : 1;
			unsigned enable : 1;
			unsigned intr   : 1;
			unsigned wrap   : 1;
		}bits;
	}s;
} rx_buffer_desc_t;

typedef struct tx_buffer_desc_t{
	unsigned int *src;
	unsigned int len;
	unsigned int *desc;
	union{
		unsigned int reg;
		struct{
			unsigned status : 16;
			unsigned res    : 12;
			unsigned full   : 1;
			unsigned last   : 1;
			unsigned intr   : 1;
			unsigned wrap   : 1;
	      	}bits;
	}s;
} tx_buffer_desc_t;

typedef struct ns9750_eth_info{
	struct sk_buff *tx_skb[NS_TX_DESC];
	struct sk_buff *rx_skb[NS_RX_DESC];

	dma_addr_t tx_dma[NS_TX_DESC];

	rx_buffer_desc_t *rx_desc;
	dma_addr_t rx_dma_desc;

	int tx_dma_next;
	int tx_dma_start;
	int tx_dma_free;

	unsigned short last_link_status;
	spinlock_t lock;

	phy_type_t phy_type;
	unsigned int phy_mdio_max_clk;

	struct net_device_stats stats;
} ns9750_eth_info;

static const u8 default_mac[6]	= {0x00, 0x00, 0x00, 0x00, 0x00, 0x00};


#include "../../i2c/chips/at24cxx.h"
extern int eeprom_ioctl(struct inode *inode, struct file *file,
                        unsigned int cmd, unsigned long arg);
#define GET_MAC_ADDR(pD, pMac)                                          \
        {                                                               \
                struct eeprom_s eeprom;                                 \
                                                                        \
                eeprom.addr = 0x0;                                      \
                eeprom.count = ETH_ALEN;                                \
                eeprom.data = pMac;                                     \
                eeprom_ioctl                                            \
                  (NULL, NULL, GET_EEPROM_DATA, (unsigned long)&eeprom);\
        }



static u32
ns_mii_poll_busy(void)
{
	u32 timeout = NS_MII_POLE_TIMEOUT;

	while((RegRd32(NS_ETH_MIND) & NS_ETH_MIND_BUSY) && timeout--);
	
	return timeout;
}

static u16
ns_mii_read(u16 reg)
{
	RegWr32(NS_ETH_MADR, (NS_ETH_PHY_ADDRESS << 8) | reg);
	RegWr32(NS_ETH_MCMD, NS_ETH_MCMD_READ);
	
	if(!ns_mii_poll_busy())
		PRINT_WARN("MII still busy in read\n");
	
	RegWr32(NS_ETH_MCMD, 0);
	
	return (u16)RegRd32(NS_ETH_MRDD);
}

static void
ns_mii_write(u16 reg, u16 data)
{
	RegWr32(NS_ETH_MADR, (NS_ETH_PHY_ADDRESS << 8) | reg);
	RegWr32(NS_ETH_MWTD, data);
	
	if(!ns_mii_poll_busy())
		PRINT_WARN("MII still busy in write\n");
}

static u32
ns_mii_identify_phy(struct net_device *pdev)
{
	struct ns9750_eth_info *pinfo = PRIV(pdev);
	char *name = NULL;
	u32 phy_id;
	u16 phy_id1, phy_id2;
	
	phy_id1 = ns_mii_read(PHY_COMMON_ID1);
	phy_id2 = ns_mii_read(PHY_COMMON_ID2);
	
	phy_id = ((phy_id1 << 16) | phy_id2) & PHY_ID_MASK;

	switch(phy_id){
	case PHY_LXT971A:
		name = "LXT971A/972A";
		pinfo->phy_mdio_max_clk = PHY_LXT971_MDIO_MAX_CLK;
		break;
	case PHY_ICS1893BK:
		name = "ICS1893BK";
		pinfo->phy_mdio_max_clk = PHY_ICS1893_MDIO_MAX_CLK;
		break;
	case PHY_NONE:
	default:
		phy_id = 0;
		name = "unknown";
		pinfo->phy_mdio_max_clk = PHY_MDIO_MAX_CLK;
		break;
	}
	
	PRINT_INFO("PHY (0x%x, 0x%x) = %s detected at 0x%x\n",
		   phy_id1, phy_id2, name, NS_ETH_PHY_ADDRESS);

	return phy_id;
}

static u32
ns_mii_get_clock_divisor(u32 max_mdio_clk)
{
	struct {
		u32 clk_div;
		u32 clks;
	} phy_clock_divisors[] = {
		{  4, NS_ETH_MCFG_CLKS_4 },
		{  6, NS_ETH_MCFG_CLKS_6 },
		{  8, NS_ETH_MCFG_CLKS_8 },
		{ 10, NS_ETH_MCFG_CLKS_10 },
		{ 20, NS_ETH_MCFG_CLKS_20 },
		{ 30, NS_ETH_MCFG_CLKS_30 },
		{ 40, NS_ETH_MCFG_CLKS_40 },
	};

	int i;
	int size = sizeof( phy_clock_divisors ) / \
		   sizeof( phy_clock_divisors[0] );
	u32 clks = NS_ETH_MCFG_CLKS_40;

	for(i=0; i<size; i++){
		if((AHB_CLK_FREQ / phy_clock_divisors[i].clk_div) <= \
		   max_mdio_clk){
			clks = phy_clock_divisors[i].clks;
			break;
		}
	}
	
	return clks;
}

static void
ns_link_print_changed(struct net_device* pdev)
{
	struct ns9750_eth_info *pinfo = PRIV(pdev);
	u16 status;
	u16 control;

	control = ns_mii_read(PHY_COMMON_CTRL);

	if((control & PHY_COMMON_CTRL_AUTO_NEG) == PHY_COMMON_CTRL_AUTO_NEG){
		/* PHY_COMMON_STAT_LNK_STAT is only set on autonegotiation */
		status = ns_mii_read(PHY_COMMON_STAT);

		if(!(status & PHY_COMMON_STAT_LNK_STAT)){
			netif_carrier_off(pdev);
			return;
		}else{
			if(!netif_carrier_ok(pdev))
				/* we are online again, enable it */
				netif_carrier_on(pdev);

			if(pinfo->phy_type == PHY_LXT971A){
				status = ns_mii_read(PHY_LXT971_STAT2);
				status &= (PHY_LXT971_STAT2_100BTX |
					   PHY_LXT971_STAT2_DUPLEX_MODE |
					   PHY_LXT971_STAT2_AUTO_NEG);

				/* mask out all uninteresting parts */
			}
			/* other PHYs must store there link information in
			   uiStatus as PHY_LXT971 */
		}
	}else{
		/* mode has been forced, so uiStatus should be the same as the
		   last link status, enforce printing */
		status = pinfo->last_link_status;
		pinfo->last_link_status = LINK_STATUS_UNKNOWN;
	}

	if(status != pinfo->last_link_status){
		/* save current link status */
		pinfo->last_link_status = status;

		/* print new link status */
		DEBUG_INFO("link mode %i Mbps %s duplex %s\n",
		     (status & PHY_LXT971_STAT2_100BTX) ? 100 : 10,
		     (status & PHY_LXT971_STAT2_DUPLEX_MODE) ? "full" : "half",
		     (status & PHY_LXT971_STAT2_AUTO_NEG) ? "(auto)" : "");
	}
}

static void
ns_link_update_egcr(struct net_device* pdev)
{
	struct ns9750_eth_info *pinfo = PRIV(pdev);
	u32 egcr;
	u32 mac2;
	u32 ipgt;

	egcr = RegRd32(NS_ETH_EGCR1);
	mac2 = RegRd32(NS_ETH_MAC2);
	ipgt = RegRd32(NS_ETH_IPGT) & ~NS_ETH_IPGT_MA;

	mac2 &= ~NS_ETH_MAC2_FULLD;
	if((pinfo->last_link_status & PHY_LXT971_STAT2_DUPLEX_MODE)
	    == PHY_LXT971_STAT2_DUPLEX_MODE){
		mac2 |= NS_ETH_MAC2_FULLD;
		ipgt |= 0x15;	/* see [1] p. 339 */
	}else{
		ipgt |= 0x12;	/* see [1] p. 339 */
	}

	RegWr32(NS_ETH_MAC2, mac2);
	RegWr32(NS_ETH_EGCR1, egcr);
	RegWr32(NS_ETH_IPGT, ipgt);
}

#if 0 /* unused */
static void
ns_link_force(struct net_device* pdev)
{
	struct ns9750_eth_info *pinfo = PRIV(pdev);
	u16 control;

	control = ns_mii_read(PHY_COMMON_CTRL);
	control &= ~(PHY_COMMON_CTRL_SPD_MA |
		     PHY_COMMON_CTRL_AUTO_NEG |
		     PHY_COMMON_CTRL_DUPLEX);

	pinfo->last_link_status = 0;

	control |= PHY_COMMON_CTRL_SPD_10;

	ns_mii_write(PHY_COMMON_CTRL, control);

	ns_link_print_changed(pdev);
	ns_link_update_egcr(pdev);
}
#endif

static void 
ns_link_auto_negotiate(struct net_device *pdev)
{
  	unsigned long start_jiffies;
	u16 status;

	DEBUG_FUNC();

	status = ns_mii_read(PHY_COMMON_STAT);
	if((status & (PHY_COMMON_STAT_AN_COMP | PHY_COMMON_STAT_LNK_STAT)) ==
	   (PHY_COMMON_STAT_AN_COMP | PHY_COMMON_STAT_LNK_STAT)){
		ns_link_print_changed(pdev);
		ns_link_update_egcr(pdev);
		return;
	}

	/* run auto-negotation */
	/* define what we are capable of */
	ns_mii_write(PHY_COMMON_AUTO_ADV,
		     PHY_COMMON_AUTO_ADV_100BTXFD |
		     PHY_COMMON_AUTO_ADV_100BTX |
		     PHY_COMMON_AUTO_ADV_10BTFD |
		     PHY_COMMON_AUTO_ADV_10BT |
		     PHY_COMMON_AUTO_ADV_802_3);
	/* start auto-negotiation */
	ns_mii_write(PHY_COMMON_CTRL,
		     PHY_COMMON_CTRL_AUTO_NEG |
		     PHY_COMMON_CTRL_RES_AUTO);

	/* wait for completion */
	start_jiffies = jiffies;
	while(jiffies < start_jiffies + NS_MII_NEG_DELAY){
		status = ns_mii_read(PHY_COMMON_STAT);
		if((status & 
		    (PHY_COMMON_STAT_AN_COMP | PHY_COMMON_STAT_LNK_STAT)) ==
		   (PHY_COMMON_STAT_AN_COMP | PHY_COMMON_STAT_LNK_STAT)){
			/* lucky we are, auto-negotiation succeeded */
			ns_link_print_changed(pdev);
			ns_link_update_egcr(pdev);
			return;
		}
	}

	//PRINT_WARN("auto-negotiation timed out, forcing 10Mbps/Half\n");

	//ns_link_force(pdev);
}

static int
ns_link_init(struct net_device *pdev)
{
        u16 val;

	DEBUG_FUNC();

	ns_link_auto_negotiate(pdev);

        val = ns_mii_read(PHY_COMMON_STAT);

        if(val & PHY_COMMON_STAT_LNK_STAT){
                return 0;
        }
	return -1;
}



//static inline void 
static void 
ns9750_eth_rx_process(struct net_device *pdev, int ring)
{
	struct ns9750_eth_info *pinfo = PRIV(pdev);
	int i;
	int j;
	volatile int k;
	rx_buffer_desc_t* desc;

	desc = &pinfo->rx_desc[ring * NS_RX_DESC_RING];
  
	for(i = ring * NS_RX_DESC_RING; i < (ring + 1) * NS_RX_DESC_RING; i++){
	  if(desc->s.bits.full){
	    if(desc->s.bits.status & NS_ETH_ERSR_RXOK){
	      pinfo->stats.rx_packets++;
	      pinfo->stats.rx_bytes += desc->len;
	
	      skb_put(pinfo->rx_skb[i],desc->len);
	      
	      pinfo->rx_skb[i]->protocol = 
	      	eth_type_trans(pinfo->rx_skb[i], pdev);
	      pinfo->rx_skb[i]->ip_summed = CHECKSUM_UNNECESSARY;
	
	      consistent_sync(pinfo->rx_skb[i]->data, desc->len, 
			      DMA_FROM_DEVICE);
	
	      /* This loop is not for cache updates but helps
	       * "breaking" a resource conflict of the ethernet engine
	       * and the CPU. If not present, when receiving many
	       * frames (e.g. nfsmount by initramfs on 2.6.8, we'll
	       * get an error message
	       * ns_eth_rx_int: rx error status: 0x2010000
	       * (RX FIFO overflow). Adjusting AHB Arbiter BRC
	       * Registers (with either CPU or Ethernet more priority)
	       * doesn't help. A udelay(10) at this place would have
	       * the same effect but would run longer I guess.*/
	      for(j = 0; j < 3; j++)
		k = ((int*)pinfo->rx_skb[i]->data)[j];

#ifdef NS9750_RXALIGN
	      skb_reserve(pinfo->rx_skb[i], 2); /* align 16byte */
#endif
	
	      netif_rx(pinfo->rx_skb[i]);
	
	      pdev->last_rx = jiffies;
	
	      pinfo->rx_skb[i] = dev_alloc_skb(NS_ETH_FRAME_LEN + 2);
	
	      if(pinfo->rx_skb[i] != NULL){
		pinfo->rx_skb[i]->dev = pdev;
	  
		/* prepare descriptor */
		pinfo->rx_desc[i].src = (unsigned int*)
		  virt_to_dma(NULL, pinfo->rx_skb[i]->data);
		pinfo->rx_desc[i].len = NS_ETH_FRAME_LEN;
		pinfo->rx_desc[i].s.bits.wrap =
		  !((i + 1) % (NS_RX_DESC_RING)) ? 1 : 0;
		pinfo->rx_desc[i].s.bits.intr = 1;
		pinfo->rx_desc[i].s.bits.enable = 1;
		pinfo->rx_desc[i].s.bits.full = 0;
	      }else{
		PRINT_ERR("out of memory\n");
		pinfo->rx_desc[i].s.bits.enable = 0;
	      }
	    }else{
	      pinfo->stats.rx_errors++;
	      if(desc->s.bits.status & NS_ETH_ERSR_RXDR)
		pinfo->stats.rx_crc_errors++;
	      if(desc->s.bits.status & NS_ETH_ERSR_RXSHT)
		pinfo->stats.rx_frame_errors++;
	    }
	  }
	  desc++;
	}
}

  
static irqreturn_t
ns9750_eth_rx_interrupt(int irq, void *arg, struct pt_regs *pregs)
{
	struct net_device *pdev = arg;
	struct ns9750_eth_info *pinfo = PRIV(pdev);
	u32 status;
	int ring;
	u32 mask;
	
	spin_lock(&pinfo->lock);
  
	/* acknowledge interrupt */
	status = RegRd32(NS_ETH_EINTR) & NS_ETH_EINTR_RX_MA;
	RegWr32(NS_ETH_EINTR, status);

	if(status & (NS_ETH_EINTR_RXNOBUF |
		     NS_ETH_EINTR_RXBUFFUL |
		     NS_ETH_EINTR_RXOVL_DATA |
		     NS_ETH_EINTR_RXOVL_STAT))
		PRINT_ERR("rx error status: 0x%x\n", status);

	mask = NS_ETH_EINTR_RXDONEA;
	for(ring = 0; ring < NS_RX_RINGS; ring++){
		if(status & mask){
		  //printk("ring: %d\n",ring);
			ns9750_eth_rx_process(pdev, ring);
			RegWr32(NS_ETH_RXFREE, 1 << ring);
		}
		
		mask >>= 1;
	}
	
	spin_unlock(pinfo->lock);
	
	return IRQ_HANDLED;
}

static irqreturn_t
ns9750_eth_tx_interrupt(int irq, void *arg, struct pt_regs *pregs)
{
	struct net_device *pdev = arg;
	ns9750_eth_info *pinfo = PRIV(pdev);
	u32 status, tx_status;
	u32 val;
	int i;
	tx_buffer_desc_t* tx_buffer_desc;
	char transmitted = 0;
	int first_not_transmitted = -1;
	char kick_dma = 0;
	
	spin_lock(&pinfo->lock);

	/* acknowledge interrupt */
	status = RegRd32(NS_ETH_EINTR);
	RegWr32(NS_ETH_EINTR, status & NS_ETH_EINTR_TX_MA);
	
	if(status & (NS_ETH_EINTR_TXERR | NS_ETH_EINTR_TXBUFNR))
	  PRINT_WARN("tx error status: 0x%x\n", status);
	
	/* check each frame if it has been sent */
	tx_buffer_desc = (tx_buffer_desc_t*)NS_ETH_TXBD + pinfo->tx_dma_start;
	/* frames sent */
	for(i = pinfo->tx_dma_start; i < pinfo->tx_dma_next;
	    i++, tx_buffer_desc++){
	  if(pinfo->tx_skb[i] == NULL)
	    /* has been processes in a previous round with
	     * some packets untransmitted */
	    continue;
	  
	  tx_status = tx_buffer_desc->s.reg;
	  if(!(tx_status & 0x10000000)){
	    /* we may have had untransmitted the last round */
	    if(!(tx_buffer_desc->s.reg & NS_ETH_ETSR_TXOK)){
	      PRINT_WARN("descr %i empty but not transmitted ok (0x%x)\n",
			 i, tx_buffer_desc->s.reg);
	    }
	    
	    /* buffer sent */
	    dev_kfree_skb_irq(pinfo->tx_skb[i]);
	    pinfo->tx_skb[i] = NULL;
	    transmitted = 1;
	  }else if(first_not_transmitted == -1){
	    first_not_transmitted = i;
	    /* don't iterate over known frames */
	    pinfo->tx_dma_start = i;
	    break;
	  }
	  
	  
	  if((tx_status & (NS_ETH_ETSR_TXOK | NS_ETH_ETSR_TXAL | 
			   NS_ETH_ETSR_TXAED | NS_ETH_ETSR_TXAEC | 
			   NS_ETH_ETSR_TXAUR | NS_ETH_ETSR_TXAJ)) 
	     == NS_ETH_ETSR_TXOK){
	    pinfo->stats.tx_packets++;
	    pinfo->stats.tx_bytes += tx_buffer_desc->len;
	  }else{
	    pinfo->stats.tx_errors++;
	    if(tx_status & (NS_ETH_ETSR_TXAL | NS_ETH_ETSR_TXAEC))
	      pinfo->stats.collisions++;
	    if(tx_status & NS_ETH_ETSR_TXAED)
	      pinfo->stats.tx_carrier_errors++;
	    if(tx_status & NS_ETH_ETSR_TXDEF)
	      pinfo->stats.tx_carrier_errors++;
	    if(tx_status & (NS_ETH_ETSR_TXAJ | NS_ETH_ETSR_TXCRC))
	      pinfo->stats.tx_aborted_errors++;
	  }
	}
	
	if(first_not_transmitted != -1){
	  /* retry the buggy descriptors */
	  kick_dma = 1;
	  RegWr32(NS_ETH_TXRPTR, first_not_transmitted * 4);
	}else if(transmitted){
	  if(pinfo->tx_dma_next == pinfo->tx_dma_free){
	    /* we sent all, wind descriptors back to head */
	    pinfo->tx_dma_start = pinfo->tx_dma_free = 0;
	    val = RegRd32(NS_ETH_EGCR1);
	    RegWr32(NS_ETH_EGCR1, val & ~NS_ETH_EGCR1_ETXDMA);
	  }else{
	    /* i and tx_buffer_desc point to tx_dma_next+1 */
	    pinfo->tx_dma_start = i;
	    RegWr32(NS_ETH_TXRPTR, i * 4);
	    while(i++ < pinfo->tx_dma_free){
	      tx_buffer_desc->s.reg = 0x70000000;
	      tx_buffer_desc++;
	    }
	    (tx_buffer_desc-1)->s.reg = 0xf0000000;
      
	    /* sent the next bunch of frames */
	    kick_dma = 1;
	  }
	  pinfo->tx_dma_next = pinfo->tx_dma_free;
	  
	  if(netif_queue_stopped(pdev))
	    netif_wake_queue(pdev);
	}
	
	if(kick_dma){
	  val = RegRd32(NS_ETH_EGCR2);
	  RegWr32(NS_ETH_EGCR2, val & ~NS_ETH_EGCR2_TCLER);
	  val = RegRd32(NS_ETH_EGCR2);
	  RegWr32(NS_ETH_EGCR2, val | NS_ETH_EGCR2_TCLER);
	}
	
	spin_unlock(&pinfo->lock);
	
	return IRQ_HANDLED;
}

static irqreturn_t
ns9750_eth_phy_interrupt(int irq, void *arg, struct pt_regs *pregs)
{
	struct net_device *pdev = arg;
	struct ns9750_eth_info *pinfo = PRIV(pdev);
	u32 status = 0;

	DEBUG_FUNC();

	spin_lock(&pinfo->lock);

	/* ACK */
	if(pinfo->phy_type == PHY_LXT971A){
		status = ns_mii_read(PHY_LXT971_INT_STATUS);
		DEBUG_INFO("int status: 0x%04x\n", status);
	}

	if(status & PHY_LXT971_INT_STATUS_LINKCHG){
		pinfo->last_link_status = LINK_STATUS_UNKNOWN;
		status = ns_mii_read(PHY_COMMON_STAT);
		if(status & PHY_COMMON_STAT_LNK_STAT){
			DEBUG_INFO("Link is up!\n");
			ns_link_auto_negotiate(pdev);
		}else{
			DEBUG_INFO("Link is down!\n");
		}
	}

	spin_unlock(pinfo->lock);

	return IRQ_HANDLED;
}

static int
ns9750_eth_reset(struct net_device *pdev)
{
	struct ns9750_eth_info *pinfo = PRIV(pdev);
	u32 val;

	DEBUG_FUNC();
  
	/* Reset MAC */
	val = RegRd32(NS_ETH_EGCR1);
	RegWr32(NS_ETH_EGCR1, val | NS_ETH_EGCR1_MAC_HRST);
	udelay(5);
	RegWr32(NS_ETH_EGCR1, val & ~NS_ETH_EGCR1_MAC_HRST);

	/* Reset and Initialize PHY */
	val = RegRd32(NS_ETH_MAC1);
	RegWr32(NS_ETH_MAC1, val & ~NS_ETH_MAC1_SRST);
	
	/* Set MDIO Clock */
	RegWr32(NS_ETH_MCFG,
		ns_mii_get_clock_divisor(PHY_MDIO_MAX_CLK));
	
	/* Reset PHY */
	val = ns_mii_read(PHY_COMMON_CTRL);
	ns_mii_write(PHY_COMMON_CTRL, val | PHY_COMMON_CTRL_RESET);
	udelay(350);

#if 0	
	/* MII clock has been setup to default, ns_mii_identify_phy should
	   work for all */
	if(!(pinfo->phy_type = ns_mii_identify_phy(pdev))){
		PRINT_ERR("Unsupported PHY, aborting\n");
		return 0;
	}
#endif

	/* now take the highest MDIO clock possible after detection */
	RegWr32(NS_ETH_MCFG,
		ns_mii_get_clock_divisor(pinfo->phy_mdio_max_clk));

	pinfo->last_link_status = LINK_STATUS_UNKNOWN;

#if 0
	switch(pinfo->phy_type){
	case PHY_LXT971A:
		ns_mii_write(PHY_LXT971_LED_CFG,
			     PHY_LXT971_LED_CFG_LINK_ACT << 
			     PHY_LXT971_LED_CFG_SHIFT_LED2);
		break;
	default:
		break;
	}
#endif

	/* reset interrupt status register */
	RegWr32(NS_ETH_EINTR, NS_ETH_EINTR_RX_MA | NS_ETH_EINTR_TX_MA);

	return 1;
}

static int
ns9750_eth_setup(struct net_device *pdev)
{
	u32 val;
	
	DEBUG_FUNC();

	ns9750_eth_reset(pdev);
  
	ns_link_init(pdev);
	
	/* Set MAC Address*/
	RegWr32(NS_ETH_SA1, ((pdev->dev_addr[5] << 8) | pdev->dev_addr[4]));
	RegWr32(NS_ETH_SA2, ((pdev->dev_addr[3] << 8) | pdev->dev_addr[2]));
	RegWr32(NS_ETH_SA3, ((pdev->dev_addr[1] << 8) | pdev->dev_addr[0]));
	
	val = RegRd32(NS_ETH_SAFR);
	RegWr32(NS_ETH_SAFR, val | NS_ETH_SAFR_BROAD );

	return 0;
}

static void
ns9750_eth_update_rx_desc(struct net_device *pdev)
{
	int timeout = 20;
	u32 val;

	val = RegRd32(NS_ETH_EGCR1);
	RegWr32(NS_ETH_EGCR1, val | NS_ETH_EGCR1_ERXINIT);

	/* [1] Tab. 221 states less than 5us */
	while(--timeout && !(RegRd32(NS_ETH_EGSR) & NS_ETH_EGSR_RXINIT))
		/* wait for finish */
		udelay(1);

	if(!timeout)
		PRINT_WARN("ethernet descriptor update failed\n");

	/* @TODO do we need to clear RXINIT? */
	val = RegRd32(NS_ETH_EGCR1);
	RegWr32(NS_ETH_EGCR1, val & ~NS_ETH_EGCR1_ERXINIT);

	/* all 4 descriptors free */
	RegWr32(NS_ETH_RXFREE, 0xf);
}

static void
ns9750_eth_irq_enable(struct net_device *pdev)
{
	DEBUG_FUNC();

	enable_irq(IRQ_ETH_RX);
	enable_irq(IRQ_ETH_TX);

	enable_irq(IRQ_ETH_PHY);
#if 1
	ns_mii_write(PHY_LXT971_INT_ENABLE, PHY_LXT971_INT_ENABLE_INTEN | PHY_LXT971_INT_ENABLE_LINKMSK);
#endif
}

static void
ns9750_eth_irq_disable(struct net_device *pdev)
{
	DEBUG_FUNC();

#if 1
	ns_mii_write(PHY_LXT971_INT_ENABLE, 0);
	ns_mii_read(PHY_LXT971_INT_STATUS);
#endif

	disable_irq(IRQ_ETH_RX);
	disable_irq(IRQ_ETH_TX);
#if 1
	disable_irq(IRQ_ETH_PHY);
#endif
}

static int
ns9750_eth_enable(struct net_device *pdev)
{
	struct ns9750_eth_info *pinfo = PRIV(pdev);
	u32 val;
	int i;

	DEBUG_FUNC();

	/* enable hardware */
	/* enable receive and transmit FIFO, use 10/100 Mbps MII */
	RegWr32(NS_ETH_EGCR1,
		NS_ETH_EGCR1_ETXWM |
		NS_ETH_EGCR1_ERX |
		NS_ETH_EGCR1_ETX |
#ifdef NS9750_RXALIGN
		NS_ETH_EGCR1_RXALIGN |
#endif
		NS_ETH_EGCR1_PHY_MODE_MII |
		NS_ETH_EGCR1_ITXA);

	/* EINTR can only be reseted with ERX and ETX on */
	val = RegRd32(NS_ETH_EINTR);
	RegWr32(NS_ETH_EINTR, val);

	RegWr32(NS_ETH_MAC1, NS_ETH_MAC1_RXEN);

	/* the linux kernel may give packets < 60 bytes, for example arp */
	RegWr32(NS_ETH_MAC2,
		NS_ETH_MAC2_CRCEN | NS_ETH_MAC2_PADEN);

	/* program the first descriptors for each ring. */
	for(i = 0; i < NS_RX_RINGS; i++)
		RegWr32(NS_ETH_RXAPTR + (i * 4),
			(u32)(pinfo->rx_dma_desc + (i * NS_RX_DESC_RING)));

	ns9750_eth_update_rx_desc(pdev);

	val = RegRd32(NS_ETH_EGCR1);
	RegWr32(NS_ETH_EGCR1, val | NS_ETH_EGCR1_ERXDMA);

	/* ignore statistics interrupt */
	RegWr32(NS_ETH_EINTREN ,
		NS_ETH_EINTR_RXOVL_DATA |
		NS_ETH_EINTR_RXBUFC |
		NS_ETH_EINTR_RXDONEA |
		NS_ETH_EINTR_RXDONEB |
		NS_ETH_EINTR_RXDONEC |
		NS_ETH_EINTR_RXDONED |
		NS_ETH_EINTR_RXNOBUF |
		NS_ETH_EINTR_RXBUFFUL |
		NS_ETH_EINTR_TXBUFC |
		NS_ETH_EINTR_TXBUFNR |
		NS_ETH_EINTR_TXDONE |
		NS_ETH_EINTR_TXERR |
		NS_ETH_EINTR_TXIDLE);

	/* clear statistics */
	RegWr32(NS_ETH_EGCR2, NS_ETH_EGCR2_CLRCNT);
	/* enable statistics */
	RegWr32(NS_ETH_EGCR2, NS_ETH_EGCR2_AUTOZ | NS_ETH_EGCR2_STEN);
        return 0;
}

static int
ns9750_eth_disable(struct net_device *pdev)
{
	u32 val;
	val = RegRd32(NS_ETH_MAC1);
	RegWr32(NS_ETH_MAC1, val & ~NS_ETH_MAC1_RXEN);
	
	val = RegRd32(NS_ETH_EGCR1);
	RegWr32(NS_ETH_EGCR1, val & ~(NS_ETH_EGCR1_ERX |
				      NS_ETH_EGCR1_ERXDMA |
				      NS_ETH_EGCR1_ETX |
				      NS_ETH_EGCR1_ETXDMA));
	
	return 0;
}

static int
ns9750_eth_open(struct net_device *pdev)
{
	int status;

	DEBUG_FUNC();

	try_module_get(THIS_MODULE);

	status = ns9750_eth_setup(pdev);
	if(status != 0)
		return -EAGAIN;

	ns9750_eth_enable(pdev);
	ns9750_eth_irq_enable(pdev);
	
	netif_start_queue(pdev);
	
	return 0;
}

static int
ns9750_eth_close(struct net_device *pdev)
{
	DEBUG_FUNC();

	ns9750_eth_irq_disable(pdev);
	ns9750_eth_disable(pdev);
	netif_stop_queue(pdev);
	
	module_put(THIS_MODULE);
	
	return 0;
}

static int
ns9750_eth_hard_start_xmit(struct sk_buff *pskb, struct net_device *pdev)
{
	struct ns9750_eth_info *pinfo = PRIV(pdev);
	unsigned long flags;
	tx_buffer_desc_t* tx_buffer_desc;
	u32 val;

        spin_lock_irqsave(&pinfo->lock, flags);

	if(pinfo->tx_dma_free == NS_TX_DESC - 1){
		/* descriptor table full, wait until isr cleared them */
		/* @TODO: we could use the wraparound feature and use the
		 * descriptor table as a ring fifo to tx_start-1 */
		netif_stop_queue(pdev);
		spin_unlock_irqrestore(&pinfo->lock, flags);
		return 1;
	}

	pinfo->tx_skb[pinfo->tx_dma_free] = pskb;
	pinfo->tx_dma[pinfo->tx_dma_free] = 
		dma_map_single(NULL, pskb->data, pskb->len, DMA_TO_DEVICE);

	/* configure next free buffer descriptor */
	tx_buffer_desc = (tx_buffer_desc_t*)NS_ETH_TXBD;
	tx_buffer_desc += pinfo->tx_dma_free;

	/* skb->data is 32bit aligned */
	tx_buffer_desc->src = (unsigned int*)pinfo->tx_dma[pinfo->tx_dma_free];
	tx_buffer_desc->len = (pskb->len >= ETH_ZLEN ? pskb->len : ETH_ZLEN);

	/* only 32bit accesses allowed.wrap, last, full  */
	tx_buffer_desc->s.reg = 0x00000000;

	if(!pinfo->tx_dma_free){
		/* no packages have been added while doing a transfer, begin at
		 * start of descriptor table */
		pinfo->tx_dma_next = pinfo->tx_dma_free + 1;
		tx_buffer_desc->s.reg |= 0xf0000000;

		/* enable DMA engine to transmit descriptor */

		RegWr32(NS_ETH_TXPTR,0);
		RegWr32(NS_ETH_TXRPTR, 0 );
		val = RegRd32(NS_ETH_EGCR1);
		RegWr32(NS_ETH_EGCR1, val | NS_ETH_EGCR1_ETXDMA);
		val = RegRd32(NS_ETH_EGCR2);
		RegWr32(NS_ETH_EGCR2, val & ~NS_ETH_EGCR2_TCLER);
		val = RegRd32(NS_ETH_EGCR2);
		RegWr32(NS_ETH_EGCR2, val | NS_ETH_EGCR2_TCLER);
	}

	pdev->trans_start = jiffies;

	pinfo->tx_dma_free++;

        spin_unlock_irqrestore(&pinfo->lock, flags);

	return 0;
}

static void 
ns9750_eth_tx_timeout(struct net_device *pdev)
{
	struct ns9750_eth_info *pinfo = PRIV(pdev);

	DEBUG_FUNC();

	pinfo->stats.tx_errors++;
	
        if(pinfo->last_link_status == LINK_STATUS_UNKNOWN){
		ns_link_init(pdev);
        }
}

static struct net_device_stats *
ns9750_eth_get_stats(struct net_device *pdev)
{
	DEBUG_FUNC();
	return &PRIV(pdev)->stats;
}

static u16
ns9750_rotate(u16 reg, int direction, int bits)
{
	int i;

	if(direction == NS_LEFT){
		for(i = 0; i < bits; i++){
			if(reg & 0x8000) /* if high bit set */
				/* shift left and or in 1 */
				reg = (reg << 1) | 1; 
			else
				reg <<= 1;
		}
	}else{
		for(i = 0; i < bits; i++){
			if(reg & 1) /* if low bit set */
				/* shift right and or in hi bit */
				reg = (reg>>1)|0x8000;
			else
				reg >>= 1;
		}
	}

	return reg;
}

static int 
ns9750_calculate_hash_bit(u8 *mac_addr)
{
	u16 copy_mac[3];
	u16 *mac;
	u16 bp;
	u16 bx;
	unsigned long crc;
	int mac_word;
	int bit_index;

	/* mac may not be 32 bit aligned */
	memcpy(copy_mac, mac_addr, sizeof(copy_mac));

	mac = copy_mac;
	crc = 0xffffffffL;

	for(mac_word = 0; mac_word < 3; mac_word++){
		bp = *mac;
		mac++;
		
		for(bit_index = 0; bit_index < 16; bit_index++){
			bx = (u16)(crc >> 16);
			bx = ns9750_rotate(bx,NS_LEFT,1);
			bx ^= bp;
			crc <<= 1;
			bx &= 1;
			if(bx)
				crc ^= NS_POLYNOMIAL;
			crc |= bx;
			bp = ns9750_rotate(bp, NS_RIGHT, 1);
		}
	}

	return (crc >> 23) & 0x3f;
}

static void 
ns9750_set_hash_bit(u8 *hash_table, int bit)
{
	int byte_index;
	int bit_index;

	byte_index = bit >> 3;
	bit_index = bit & 7;
	hash_table[byte_index] |= (1 << bit_index);
}

static void 
ns9750_load_mca_hash_table(struct dev_mc_list* plist)
{
	unsigned long hash_table[2];
	struct dev_mc_list* pentry;
	int bit;

	/* Generate the hash table. */
	memset(hash_table, 0, sizeof(hash_table));

	for(pentry = plist; pentry != NULL; pentry = pentry->next){
		bit = ns9750_calculate_hash_bit(pentry->dmi_addr);
		ns9750_set_hash_bit((u8 *) hash_table, bit);
	}

	/* Load the hash table into the Ethernet controller. */
	RegWr32(NS_ETH_HT1, hash_table[0]);
	RegWr32(NS_ETH_HT2, hash_table[1]);
}

static void
ns9750_eth_set_multicast_list(struct net_device *pdev)
{
	u32 val;

	DEBUG_FUNC();

	val = RegRd32(NS_ETH_SAFR);

	/* disable all multicast modes */
	val &= ~(NS_ETH_SAFR_PRO | NS_ETH_SAFR_PRM | NS_ETH_SAFR_PRA);

	if(pdev->flags & IFF_PROMISC){
		RegWr32(NS_ETH_SAFR, val | NS_ETH_SAFR_PRO);
		return;
	}else if(pdev->flags & IFF_ALLMULTI){
		RegWr32(NS_ETH_SAFR,
			val | NS_ETH_SAFR_BROAD | NS_ETH_SAFR_PRM);
		return;
	}else if(!(pdev->flags & IFF_MULTICAST)){
		RegWr32(NS_ETH_SAFR, val | NS_ETH_SAFR_BROAD);
		return;
	}

	/* store multicast list of kernel into chip */
	ns9750_load_mca_hash_table(pdev->mc_list);
	RegWr32(NS_ETH_SAFR, val | NS_ETH_SAFR_BROAD | NS_ETH_SAFR_PRA);

	return;
}

static int
ns9750_eth_queue_init(struct net_device *pdev)
{
	struct ns9750_eth_info *pinfo = PRIV(pdev);
	int i;

	DEBUG_FUNC();

	pinfo->rx_desc = (rx_buffer_desc_t*)
		dma_alloc_coherent(NULL,
				   sizeof(rx_buffer_desc_t) * NS_RX_DESC,
				   &pinfo->rx_dma_desc, GFP_KERNEL | GFP_DMA);
	if(pinfo->rx_desc == NULL){
		/* ENOMEM */
		return -1;
	}

	/* preallocate any receive skb buffers for DMA transfer
	   we receive the rx interrupt after packets have been copied! */
	for(i = 0; i < NS_RX_DESC; i++){
		pinfo->rx_skb[i] = dev_alloc_skb(NS_ETH_FRAME_LEN + 2);
		if(pinfo->rx_skb[i] == NULL){
			/* ENOMEM */
			return -1;
		}
		pinfo->rx_skb[i]->dev = pdev;

		/* prepare descriptor */
		pinfo->rx_desc[i].src =
			(u32 *)virt_to_dma(NULL, pinfo->rx_skb[i]->data);

		pinfo->rx_desc[i].len = NS_ETH_FRAME_LEN;
		/* mark only the last desc in the ring to be wrap */
		pinfo->rx_desc[i].s.bits.wrap = 
			!((i + 1) % (NS_RX_DESC_RING)) ? 1 : 0;
		pinfo->rx_desc[i].s.bits.intr = 0;
		pinfo->rx_desc[i].s.bits.enable = 1;
		pinfo->rx_desc[i].s.bits.full = 0;
	}
	return 0;
}

int __init
ns9750_eth_probe(struct net_device *pdev)
{
	struct ns9750_eth_info *pinfo;
	int err;
	int i;

	DEBUG_FUNC();
	
	if(PRIV(pdev) == 0){
		pdev->priv = kmalloc(sizeof(struct ns9750_eth_info),
				     GFP_KERNEL);
		if(PRIV(pdev) == 0)
			return -ENOMEM;
	}
	memset(PRIV(pdev), 0, sizeof(struct ns9750_eth_info));
	
	pinfo = PRIV(pdev);

	for(i=0; i<6;i++)
		pdev->dev_addr[i] = default_mac[i];

	GET_MAC_ADDR(pdev, pdev->dev_addr);

	err = (int)request_irq(pdev->irq + IRQ_ETH_RX_OFFSET,
			       ns9750_eth_rx_interrupt, 0,
			       DRIVER_NAME " receive", pdev);
	if(err){
		kfree(PRIV(pdev));
		return -EAGAIN;
	}
	err = (int)request_irq(pdev->irq + IRQ_ETH_TX_OFFSET,
			       ns9750_eth_tx_interrupt, 0,
			       DRIVER_NAME " transmit", pdev);
	if(err){
		kfree(PRIV(pdev));
		return -EAGAIN;
	}

	if(1){
	err = (int)request_irq(pdev->irq + IRQ_ETH_PHY_OFFSET,
			       ns9750_eth_phy_interrupt, 0,
			       DRIVER_NAME " phy", pdev);
	if(err){
		kfree(PRIV(pdev));
		return -EAGAIN;
	}
	}


	ns9750_eth_irq_disable(pdev);

	spin_lock_init(pinfo->lock);
	
	pdev->open			= ns9750_eth_open;
	pdev->stop			= ns9750_eth_close;
	pdev->hard_start_xmit		= ns9750_eth_hard_start_xmit;
	pdev->tx_timeout		= ns9750_eth_tx_timeout;
	pdev->watchdog_timeo		= HZ * 5;
	pdev->get_stats			= ns9750_eth_get_stats;
	pdev->set_multicast_list	= ns9750_eth_set_multicast_list;

	ether_setup(pdev);
  
	ns9750_eth_queue_init(pdev);
	ns9750_eth_reset(pdev);

	/* MII clock has been setup to default, ns_mii_identify_phy should
	   work for all */
	if(!(pinfo->phy_type = ns_mii_identify_phy(pdev))){
		PRINT_ERR("Unsupported PHY, aborting\n");
		return 0;
	}

#if 0 //test forever-loop
	while(1){
	  ns_mii_write(PHY_LXT971_INT_ENABLE, PHY_LXT971_INT_ENABLE_INTEN | PHY_LXT971_INT_ENABLE_LINKMSK);
	  mdelay(1000);
		       ns9750_eth_phy_interrupt(0, pdev, 0);
	}
#endif

	if(0){
	err = (int)request_irq(pdev->irq + IRQ_ETH_PHY_OFFSET,
			       ns9750_eth_phy_interrupt, 0,
			       DRIVER_NAME " phy", pdev);
	if(err){
		kfree(PRIV(pdev));
		return -EAGAIN;
	}
	}



	err = register_netdev(pdev);
	if(err){
		free_irq(pdev->irq + IRQ_ETH_RX_OFFSET, pdev);
		free_irq(pdev->irq + IRQ_ETH_TX_OFFSET, pdev);
		free_irq(pdev->irq + IRQ_ETH_PHY_OFFSET, pdev);
		kfree(pinfo);
		return err;
	}
	
	return 0;
}


static int
ns9750_drv_probe(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct net_device *ndev;
	struct resource *res;
	int ret;
	
	DEBUG_FUNC();
	
	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if(!res)
		return -ENODEV;
	
	if(!request_mem_region(res->start, 0x10000, DEVICE_NAME))
		return -EBUSY;
	
	ndev = alloc_etherdev(sizeof(struct ns9750_eth_info));
	if(!ndev){
		release_mem_region(res->start, 0x10000);
		return -ENOMEM;
	}

	SET_MODULE_OWNER(ndev);
	SET_NETDEV_DEV(ndev, dev);
	
	ndev->base_addr = NS_ETH_MODULE_BASE;
	ndev->irq = platform_get_irq(pdev, 0);
	
	dev_set_drvdata(dev, ndev);
	
	ret = ns9750_eth_probe(ndev);
	if(ret){
		free_netdev(ndev);
		release_mem_region(res->start, 0x10000);
		return ret;
	}
	
	return 0;
}

static int
ns9750_drv_remove(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct net_device *ndev = dev_get_drvdata(dev);
	struct resource *res;
	
	DEBUG_FUNC();
	
	dev_set_drvdata(dev, NULL);
	
	unregister_netdev(ndev);
	
	free_irq(ndev->irq, ndev);
	
	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	release_mem_region(res->start, 0x10000);
	
	free_netdev(ndev);
	
	return 0;
}

static struct device_driver ns9750_eth_driver = {
	.name	= DEVICE_NAME,
	.bus	= &platform_bus_type,
	.probe	= ns9750_drv_probe,
	.remove	= ns9750_drv_remove,
};

static int __init
ns9750_eth_init(void)
{
	DEBUG_FUNC();
	printk("NS9750 ethernet driver.\n");
	return driver_register(&ns9750_eth_driver);
}

static void __exit
ns9750_eth_cleanup(void)
{
	DEBUG_FUNC();
	driver_unregister(&ns9750_eth_driver);
}

module_init(ns9750_eth_init);
module_exit(ns9750_eth_cleanup);

MODULE_AUTHOR("Atmark Techno, Inc.");
MODULE_DESCRIPTION("NS9750 Ethernet Driver");
MODULE_LICENSE("GPL");

