/***********************************************************************
 *
 * Copyright (C) 2004 by FS Forth-Systeme GmbH.
 * All rights reserved.
 *
 * $Id: serial_ns9750.c,v 1.6 2005/10/07 10:54:28 jdietsch Exp $
 * @Author: Markus Pietrek
 * @References: [1] NS9750 Hardware Reference/December 2003
 *              [2] sa1100.c
 *              [3] 8250.c
 * @TODO: HW flow control characters not used.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 ***********************************************************************/
/***********************************************************************
 *  @History:
 *     2005/05/06 : fixed bug #396
 *     2004/08/20 : ported to kernel 2.6.8
 *     2004/06/09  : serial port is never disabled (in startup/shutdown)
 *                   using default termios setting of console
 ***********************************************************************/

#include <linux/config.h>
#include <linux/init.h>		/* __init */

#include <linux/module.h>
#include <linux/tty.h>
#include <linux/serial.h>
#include <linux/console.h>	/* struct console */
#include <linux/ioport.h>	/* request_mem_region */

#include <linux/serial_core.h>	/* uart_console_device */

#include <asm/io.h>		/* writel */
#include <asm/delay.h>		/* udelay */

#include <asm/arch/ns9750_ser.h> /* NS9750 serial registers */
#include <asm/arch/ns9750_bbus.h> /* set_gpio_cfg */
#include <asm/arch/ns9750_sys.h>
#include <asm/arch/hardware.h>	/* BBUS_CLK_FREQ */
#include <asm/arch/irqs.h>	/* IRQ_BBUS_S1RXE */

#if defined(CONFIG_MACH_ARMADILLO300)
#define NR_PORTS	2
#else
#define NR_PORTS	4
#endif

#define DRIVER_NAME	"ns_uart"

#define GPIO_TX		0
#define GPIO_RX		1
#define GPIO_RTS	2
#define GPIO_CTS	3
#define GPIO_DTR	4
#define GPIO_DSR	5
#define GPIO_RI		6
#define GPIO_DCD	7
#define GPIO_LAST	7

#ifdef DEBUG
# define DEBUG_INIT		0x0001
# define DEBUG_MINOR		0x0002
# define DEBUG_ERROR_CRIT	0x8000

static int debug_level = DEBUG_ERROR_CRIT|DEBUG_INIT;

# define DEBUG_FN( FLG ) if( (debug_level & (FLG)) == (FLG) ) \
		printk("\r%s:line %d\n", __FUNCTION__, __LINE__);
#else /* DEBUG */
# define DEBUG_FN( n )
#endif /* DEBUG */

#define USE_XTAL_CLK
//#define USE_BBUS_CLK

typedef struct
{
	struct uart_port       	port;
	int 		        hw_port;/* serial port to use. -1 for disabled */
	char 		        hw_handshake;/* 1 for enabled, 0 for disabled */
	spinlock_t		lock; /* for ser_tx_chars */
	const char*	       	gpios;
} ns_ser_port;

ns_ser_port ns_ser_ports[ NR_PORTS ] = {
	{ hw_port: -1 },
	{ hw_port: -1 },
#if !defined(CONFIG_MACH_ARMADILLO300)
	{ hw_port: -1 },
	{ hw_port: -1 }
#endif
};

static int __init ns_ser_init( void );
static void __init ns_ser_init_ports( int hw_port );
static void __exit ns_ser_exit( void );

static unsigned int ns_ser_tx_empty( struct uart_port* port );
static void ns_ser_set_mctrl( struct uart_port* port, unsigned int mctrl );
static unsigned int ns_ser_get_mctrl( struct uart_port* port );
static void ns_ser_stop_tx( struct uart_port* port, unsigned int tty_stop );
static void ns_ser_start_tx( struct uart_port* port, unsigned int tty_start);
static void ns_ser_stop_rx( struct uart_port* port );
static void ns_ser_enable_ms( struct uart_port* port );
static void ns_ser_break_ctl( struct uart_port* port, int break_state );
static int ns_ser_startup( struct uart_port* port );
static inline void ns_ser_wait_for_fifo_empty( int index );
static inline void ns_ser_wait_for_xmitr( int index );
static void ns_ser_shutdown( struct uart_port* port );
static void ns_ser_set_termios( struct uart_port* port,
				struct termios* termios,
				struct termios* old );
static void ns_ser_pm( struct uart_port* port, unsigned int state,
			   unsigned int oldstate );
static const char* ns_ser_type( struct uart_port* port );
static void ns_ser_release_port( struct uart_port* port );
static int ns_ser_request_port( struct uart_port* port );
static void ns_ser_config_port( struct uart_port* port, int flags );
static int ns_ser_verify_port( struct uart_port* port,
			       struct serial_struct* ser );
static irqreturn_t ns_ser_tx_int( int irq, void* dev_id,
				  struct pt_regs *regs );
static irqreturn_t ns_ser_rx_int( int irq, void* dev_id,
				  struct pt_regs *regs );
static void ns_ser_rx_chars( ns_ser_port* sport, struct pt_regs* regs );
static void ns_ser_tx_chars( ns_ser_port* sport );
static inline void ns_ser_tx_char( ns_ser_port* sport, char c );
static void ns_ser_set_bitrate( int cflag, int quot, int hw_port );

static int ns_console_cbaud = B9600;

#ifdef CONFIG_SERIAL_NS9750_CONSOLE

static int __init ns_console_init( void );
#if 0
static void __init ns_console_get_options( int* baud, int* parity, int* bits,
					   int hw_port );
#endif
static int __init ns_console_setup( struct console* co, char* options );
static void ns_console_write( struct console* co, const char* s,
			      unsigned int count );

static struct uart_driver ns_ser_reg;
static struct console ns_console = {
	.name		= "ttyAM",
	.write		= ns_console_write,
	.device		= uart_console_device,
	.setup		= ns_console_setup,
	.flags		= CON_PRINTBUFFER,
	.index		= -1,
	.data		= &ns_ser_reg
};

# define SERIAL_NS_CONSOLE	&ns_console
#else
# define SERIAL_NS_CONSOLE	NULL

#endif /* CONFIG_SERIAL_NS9750_CONSOLE */

static struct uart_driver ns_ser_reg = {
	.owner			= THIS_MODULE,
	.driver_name		= DRIVER_NAME,
	.devfs_name		= "ttyAM/",
	.dev_name		= "ttyAM",
	.major			= 204,
	.minor			= 16,
	.nr			= NR_PORTS,
	.cons			= SERIAL_NS_CONSOLE
};

static struct uart_ops ns_ser_pops = {
	.tx_empty	= ns_ser_tx_empty,
	.set_mctrl	= ns_ser_set_mctrl,
	.get_mctrl	= ns_ser_get_mctrl,
	.stop_tx	= ns_ser_stop_tx,
	.start_tx	= ns_ser_start_tx,
	.stop_rx	= ns_ser_stop_rx,
	.enable_ms	= ns_ser_enable_ms,
	.break_ctl	= ns_ser_break_ctl,
	.startup	= ns_ser_startup,
	.shutdown	= ns_ser_shutdown,
	.set_termios	= ns_ser_set_termios,
	.pm		= ns_ser_pm,
	.type		= ns_ser_type,
	.release_port	= ns_ser_release_port,
	.request_port	= ns_ser_request_port,
	.config_port	= ns_ser_config_port,
	.verify_port	= ns_ser_verify_port,
};
/* order of GPIOs is GPIO_TX to GPIO_LAST, used for ns_ser_port.gpios  */
static const char ns_ser_gpios[][ GPIO_LAST + 1 ] = {
	{  0,  1,  2,  3,  4,  5,  6,  7 }, /* Port B */
	{  8,  9, 10, 11, 12, 13, 14, 15 }, /* Port A */
	{ 40, 41, 42, 43, 20, 21, 22, 23 }, /* Port C */
	{ 44, 45, 46, 47, 24, 25, 26, 27 }  /* Port D */
};

static int __init ns_ser_init( void )
{
	int ret, i;

	printk( KERN_INFO "Serial: NS9xx0 driver $Revision: 1.6 $\n" );

	/* @TODO: read ports masked from command line */
	for( i = 0; i < NR_PORTS; i++ )
		ns_ser_init_ports( i );

	ret = uart_register_driver( &ns_ser_reg );
	if( ret < 0 )
		return ret;

	ns_ser_reg.tty_driver->init_termios.c_cflag =
		(ns_ser_reg.tty_driver->init_termios.c_cflag & ~CBAUD) |
		ns_console_cbaud;

	for( i = 0; i < NR_PORTS; i++ )
		if( ns_ser_ports[ i ].hw_port != -1 ) {
			uart_add_one_port( &ns_ser_reg,
					   &ns_ser_ports[ i ].port );
			if( !(readl(addr_ser_channel(NS_SER_CTRL_A,
						     ns_ser_ports[i].hw_port)) &
			      NS_SER_CTRL_A_CE ) ) {
				writel(NS_SER_CTRL_A_CE,
				       addr_ser_channel(NS_SER_CTRL_A,
							ns_ser_ports[i].hw_port));
			}
		}

	return ret;
}

static void __init ns_ser_init_ports( int hw_port )
{
	if( ns_ser_ports[ hw_port].hw_port != -1 )
		/* already initialized */
		return;

	ns_ser_ports[ hw_port ].hw_port        = hw_port;
#if defined(USE_BBUS_CLK)
	ns_ser_ports[ hw_port ].port.uartclk   = BBUS_CLK_FREQ;
#else
	ns_ser_ports[ hw_port ].port.uartclk   = XTAL_FREQ / 2;	
#endif
	ns_ser_ports[ hw_port ].port.fifosize  = 32;
	ns_ser_ports[ hw_port ].port.line      = hw_port;
	ns_ser_ports[ hw_port ].port.iotype    = UPIO_MEM;
	ns_ser_ports[ hw_port ].port.ops       = &ns_ser_pops;
	ns_ser_ports[ hw_port ].port.membase =
		(void*) addr_ser_channel( NS_SER_CTRL_A, hw_port );
	ns_ser_ports[ hw_port ].port.mapbase =
		addr_ser_channel( NS_SER_CTRL_A, hw_port );
	ns_ser_ports[ hw_port ].port.flags = UPF_BOOT_AUTOCONF;
	/* note only rhw_port irq. tx is +1 */

	//ns_ser_ports[ hw_port ].port.irq = IRQ_BBUS_S1RXE + (2*hw_port);
	ns_ser_ports[ hw_port ].port.irq = irqno_ser_channel(hw_port);
	ns_ser_ports[ hw_port ].gpios = ns_ser_gpios[ hw_port ];
	
#ifdef CONFIG_SERIAL_NS9750_HANDSHAKE
	ns_ser_ports[ hw_port ].hw_handshake = 1;
#else
	ns_ser_ports[ hw_port ].hw_handshake = 0;
#endif
	spin_lock_init(&ns_ser_ports[hw_port].port.lock);
	spin_lock_init( &ns_ser_ports[ hw_port ].lock );
}


static void __exit ns_ser_exit( void )
{
	int i;

	for( i = 0; i < NR_PORTS; i++ )
		if( ns_ser_ports[ i ].hw_port != -1 )
			uart_remove_one_port( &ns_ser_reg,
					      &ns_ser_ports[ i ].port );
	uart_unregister_driver( &ns_ser_reg );
}

static unsigned int ns_ser_tx_empty( struct uart_port* port )
{
	ns_ser_port* sport = (ns_ser_port*) port;

	DEBUG_FN( DEBUG_MINOR );

	return( readl( addr_ser_channel( NS_SER_STAT_A, sport->hw_port) ) &
		NS_SER_STAT_A_TRDY) ? TIOCSER_TEMT : 0;
}

static void ns_ser_set_mctrl( struct uart_port* port, unsigned int mctrl )
{
	ns_ser_port* sport = (ns_ser_port*) port;
	unsigned int mcr;

	DEBUG_FN(DEBUG_INIT);

	mcr = readl( addr_ser_channel( NS_SER_CTRL_A, sport->hw_port) );
	/* printk("mctrl: 0x%X\tmcr: 0x%X\n", mctrl, mcr); */

	if (mctrl & TIOCM_RTS)
		mcr |= NS_SER_CTRL_A_RTS;
	else
		mcr &= ~NS_SER_CTRL_A_RTS;

	if (mctrl & TIOCM_DTR)
		mcr |= NS_SER_CTRL_A_DTR;
	else
		mcr &= ~NS_SER_CTRL_A_DTR;

	/* printk("mctrl: 0x%X\tmcr: 0x%X\n", mctrl, mcr); */

	writel( mcr, addr_ser_channel( NS_SER_CTRL_A, sport->hw_port) );
}

static unsigned int ns_ser_get_mctrl( struct uart_port* port )
{
	ns_ser_port* sport = (ns_ser_port*) port;
	unsigned int result = 0;
	unsigned int status;

	DEBUG_FN( DEBUG_INIT );

	status = readl( addr_ser_channel( NS_SER_STAT_A, sport->hw_port) );
	if (status & NS_SER_STAT_A_DCD)
		result |= TIOCM_CD;
	if (status & NS_SER_STAT_A_CTS)
		result |= TIOCM_CTS;
	if (status & NS_SER_STAT_A_DSR)
		result |= TIOCM_DSR;
	if (status & NS_SER_STAT_A_RI)
		result |= TIOCM_RI;

	return result;
}

static void ns_ser_stop_tx( struct uart_port* port, unsigned int tty_stop )
{
	ns_ser_port* sport = (ns_ser_port*) port;

	rmwl( addr_ser_channel( NS_SER_CTRL_A, sport->hw_port ),
	      & ~NS_SER_CTRL_A_TIC_HALF );

}

/***********************************************************************
 * @Function: ns_ser_start_tx
 * @Return: n/a
 * @Descr: enables transmit interrupt on FIFO empty
 ***********************************************************************/
static void ns_ser_start_tx( struct uart_port* port, unsigned int tty_start)
{
	ns_ser_port* sport = (ns_ser_port*) port;

	rmwl( addr_ser_channel( NS_SER_CTRL_A, sport->hw_port ),
	      | NS_SER_CTRL_A_TIC_HALF );
}

static void ns_ser_stop_rx( struct uart_port* port )
{
	ns_ser_port* sport = (ns_ser_port*) port;

	DEBUG_FN( DEBUG_INIT );

	rmwl( addr_ser_channel( NS_SER_CTRL_A, sport->hw_port ),
	      & ~NS_SER_CTRL_A_RIE_RDY & ~NS_SER_CTRL_A_RIE_CLOSED );

}

static void ns_ser_enable_ms( struct uart_port* port )
{
	ns_ser_port* sport = (ns_ser_port*) port;
	unsigned int status;

	DEBUG_FN( DEBUG_INIT );

	status = readl(addr_ser_channel(NS_SER_CTRL_A, sport->hw_port));

	/* status |= NS_SER_CTRL_A_RIC_MA; */

	writel(status, addr_ser_channel(NS_SER_CTRL_A, sport->hw_port));
}

static void ns_ser_break_ctl( struct uart_port* port, int break_state )
{
	ns_ser_port* sport = (ns_ser_port*) port;
	unsigned int brk;

	DEBUG_FN(DEBUG_INIT);

	brk = readl( addr_ser_channel( NS_SER_CTRL_A, sport->hw_port) );

	if (break_state == -1)
		brk |= NS_SER_CTRL_A_BRK;  /* Assert break */
	else
		brk &= ~NS_SER_CTRL_A_BRK;  /* Deassert break */

	writel( brk, addr_ser_channel( NS_SER_CTRL_A, sport->hw_port) );
}

/***********************************************************************
 * @Function: ns_ser_startup
 * @Return: 0 if ok otherwise != 0 if interrupts could not be registered
 * @Descr:
 ***********************************************************************/

static int ns_ser_startup( struct uart_port* port )
{
	ns_ser_port* sport = (ns_ser_port*) port;
	int ret;

	DEBUG_FN( DEBUG_INIT );

	ret = request_irq( sport->port.irq, ns_ser_rx_int, 0,
			   DRIVER_NAME " rx", sport );
	if( ret )
		return ret;

	ret = request_irq( sport->port.irq+1, ns_ser_tx_int, 0,
			   DRIVER_NAME " tx", sport );
	if( ret ) {
		free_irq( sport->port.irq, sport );

		return ret;
	}

	spin_lock( &sport->port.lock );

	if( (readl( addr_ser_channel( NS_SER_CTRL_A, sport->hw_port ) ) &
	     NS_SER_CTRL_A_CE ) )
		/* this is a console if already enabled, maybe there is still
		 * something to be printed. wait until modified */
		ns_ser_wait_for_fifo_empty( sport->hw_port );

	/* enable receive interrupts only */
	rmwl( addr_ser_channel( NS_SER_CTRL_A, sport->hw_port ),
	      | NS_SER_CTRL_A_CE |
	      NS_SER_CTRL_A_RIE_RDY |
	      NS_SER_CTRL_A_RIE_CLOSED );

	writel( NS_SER_CTRL_B_RCGT,
		addr_ser_channel( NS_SER_CTRL_B, sport->hw_port ) );

	spin_unlock( &sport->port.lock );

	/* Enable modem status interrupts */

	spin_lock_irq( &sport->port.lock );
	ns_ser_enable_ms( &sport->port );
	spin_unlock_irq( &sport->port.lock );

	/* termios settings will be configured later by set_termios */

	return 0;
}

/***********************************************************************
 * @Function: ns_ser_wait_for_fifo_empty
 * @Return: nothing
 * @Descr: doesn't lock for port accesses, must been done by calling function
 ***********************************************************************/

static inline void ns_ser_wait_for_fifo_empty( int index )
{
	while( !(readl( addr_ser_channel( NS_SER_STAT_A, index ) ) &
		 NS_SER_STAT_A_TEMPTY ) )
		/* do nothing, wait for characters in FIFO being sent */
		udelay( 1 );

	/* @TODO: until we have buffer closed again, do just a delay */
	udelay( 1000 );
}

static inline void ns_ser_wait_for_xmitr( int index )
{
	while( !(readl( addr_ser_channel( NS_SER_STAT_A, index ) ) &
		 NS_SER_STAT_A_TRDY ) )
		/* do nothing, wait for characters in FIFO being sent */
		udelay( 1 );

}

static void ns_ser_shutdown( struct uart_port* port )
{
	ns_ser_port* sport = (ns_ser_port*) port;
	unsigned long flags;

	DEBUG_FN( DEBUG_INIT );

	/* disable serial port */

	ns_ser_wait_for_fifo_empty( sport->hw_port );
	/* @TODO: printk could write in this small time to FIFO */

	spin_lock_irqsave( &sport->port.lock, flags );
	rmwl( addr_ser_channel( NS_SER_CTRL_A, sport->hw_port ),
	      & ~NS_SER_CTRL_A_RIE_RDY );
	spin_unlock_irqrestore( &sport->port.lock, flags );

	free_irq( sport->port.irq, sport );
	free_irq( sport->port.irq + 1, sport );	/* transmit */
}

static void ns_ser_set_termios( struct uart_port* port,
				struct termios* termios,
				struct termios* old )
{
	ns_ser_port* sport = (ns_ser_port*) port;
	unsigned long flags;
	unsigned int baud, quot;

	DEBUG_FN( DEBUG_INIT );

	/* Ask the core to calculate the divisor for us. */
	baud = uart_get_baud_rate( port, termios, old, 0, port->uartclk/16 );
	quot = uart_get_divisor( port, baud );
	uart_update_timeout( port, termios->c_cflag, baud );

	spin_lock( &sport->port.lock );
	ns_ser_wait_for_fifo_empty( sport->hw_port );
	spin_unlock( &sport->port.lock );

	spin_lock_irqsave( &sport->port.lock, flags );
	ns_ser_set_bitrate( termios->c_cflag, quot, sport->hw_port );

	if( UART_ENABLE_MS( &sport->port, termios->c_cflag ) )
		ns_ser_enable_ms( &sport->port );

	spin_unlock_irqrestore( &sport->port.lock, flags );
}

static void ns_ser_pm( struct uart_port* port, unsigned int state,
		       unsigned int oldstate )
{
	/* power management not supported yet */
}

static const char* ns_ser_type( struct uart_port* port )
{
	return "NS9xx0";
}

static void ns_ser_release_port( struct uart_port* port )
{
	ns_ser_port* sport = (ns_ser_port*) port;
	int i;

	DEBUG_FN( DEBUG_INIT );

	/* release all gpios  */

	for( i = 0; i < GPIO_LAST; i++ )
		if( i <= GPIO_RX || sport->hw_handshake )
			release_gpio( sport->gpios[ i ], 1 );

	/* release area in ser_module */
	release_mem_region( sport->port.mapbase, 0x40 );
}

static int ns_ser_request_port( struct uart_port* port )
{
	ns_ser_port* sport = (ns_ser_port*) port;
	int ret = 0;
	int i;

	DEBUG_FN( DEBUG_INIT );

	/* request gpios */

	for( i = 0; i < GPIO_LAST; i++ )
		if( i <= GPIO_RX || sport->hw_handshake )
			if(request_gpio(sport->gpios[i],1,DRIVER_NAME)==NULL) {
				printk(	KERN_ERR DRIVER_NAME
					":GPIO %i reserved\n", sport->gpios[i] );
				ret = -EBUSY;
				goto error;
			}

	/* request area in ser_module */

	if( request_mem_region( sport->port.mapbase,
				0x40,
				DRIVER_NAME ) == NULL) {
		printk( KERN_ERR DRIVER_NAME ":request_mem_region failed\n" );
		ret = -EBUSY;
		goto error;
	}

	/* configure GPIOS. we need at least tx and rx */

	set_gpio_cfg( sport->gpios[ GPIO_TX ], NS_GPIO_CFG_FUNC_0 );
	set_gpio_cfg( sport->gpios[ GPIO_RX ], NS_GPIO_CFG_FUNC_0 );

	if( sport->hw_handshake ) {
		set_gpio_cfg( sport->gpios[ GPIO_RTS ], NS_GPIO_CFG_FUNC_0 );
		set_gpio_cfg( sport->gpios[ GPIO_CTS ], NS_GPIO_CFG_FUNC_0 );
		set_gpio_cfg( sport->gpios[ GPIO_DTR ], NS_GPIO_CFG_FUNC_0 );
		set_gpio_cfg( sport->gpios[ GPIO_DSR ], NS_GPIO_CFG_FUNC_0 );
		set_gpio_cfg( sport->gpios[ GPIO_RI  ], NS_GPIO_CFG_FUNC_0 );
		set_gpio_cfg( sport->gpios[ GPIO_DCD ], NS_GPIO_CFG_FUNC_0 );
	}

	return ret;

error:
	/* i is +1 of the last reserved gpio */
	for( i--; i > 0; i-- )
		release_gpio( sport->gpios[ i ], 1 );

	return ret;
}

static void ns_ser_config_port( struct uart_port* port, int flags )
{
	ns_ser_port* sport = (ns_ser_port*) port;

	DEBUG_FN( DEBUG_INIT );

	if( flags & UART_CONFIG_TYPE && ns_ser_request_port(&sport->port)==0)
		sport->port.type = PORT_NS9750;
}

static int ns_ser_verify_port( struct uart_port* port,
				   struct serial_struct* ser )
{
	ns_ser_port* sport = (ns_ser_port*) port;
	int ret = 0;

	DEBUG_FN( DEBUG_INIT );

	if( ser->type != PORT_UNKNOWN && ser->type != PORT_NS9750 )
		ret = -EINVAL;
	if( sport->port.irq != ser->irq )
		ret = -EINVAL;
	if( ser->io_type != UPIO_MEM )
		ret = -EINVAL;

	return ret;
}

static irqreturn_t ns_ser_tx_int( int irq, void* dev_id,
				  struct pt_regs *regs )
{
	ns_ser_port* sport = (ns_ser_port*) dev_id;

	spin_lock( &sport->lock );
	ns_ser_tx_chars( sport );
	spin_unlock( &sport->lock );

	return IRQ_HANDLED;
}

static irqreturn_t ns_ser_rx_int( int irq, void* dev_id,
				  struct pt_regs *regs )
{
	ns_ser_port* sport = (ns_ser_port*) dev_id;

	spin_lock( &sport->lock );
	ns_ser_rx_chars( sport, regs );
	spin_unlock( &sport->lock );

	return IRQ_HANDLED;
}

/***********************************************************************
 * @Function: ns_ser_rx_chars
 * @Return: n/a
 * @Descr: receives all characters from serial FIFO and stores them in tty. Is
 *         called by interrupt handler.
 ***********************************************************************/

static void ns_ser_rx_chars( ns_ser_port* sport, struct pt_regs* regs )
{
	unsigned int stat;
	unsigned int available;
	unsigned int char_cache;
	char c;
	unsigned int flag;
	struct tty_struct *tty = sport->port.info->tty;

	stat = readl( addr_ser_channel( NS_SER_STAT_A,
					sport->hw_port ) );

	stat |= NS_SER_STAT_A_RRDY; /* @TODO: check */

	while( 1 ) {
		/* read 0-4 characters */
		if ( ! (stat & NS_SER_STAT_A_RRDY ) )
			/* no more characters, return */
			break;

		available = (stat & NS_SER_STAT_A_RXFDB_MA) >> 20;
		if ( !available )
			available = 4;

		char_cache = readl( addr_ser_channel( NS_SER_FIFO,
						      sport->hw_port ) );

		flag = TTY_NORMAL;
		if ( stat & NS_SER_STAT_A_ROVER ) {
			sport->port.icount.overrun++;
			flag = TTY_OVERRUN;
		} else if( stat & NS_SER_STAT_A_RFE ) {
			sport->port.icount.frame++;
			flag = TTY_FRAME;
		} else if( stat & NS_SER_STAT_A_RPE ) {
			sport->port.icount.parity++;
			flag = TTY_PARITY;
		} else if( stat & NS_SER_STAT_A_RBRK ){
			sport->port.icount.brk++;
			flag = TTY_BREAK;
		}

		/* characters read from FIFO, give them kernel */

		while( available ) {
			if( unlikely( tty->flip.count >= TTY_FLIPBUF_SIZE ) ) {
				tty->flip.work.func( (void *) tty );
				if (tty->flip.count >= TTY_FLIPBUF_SIZE )
					break;
			}

			available--;
			c = char_cache & 0xff;
			char_cache >>= 8;

			sport->port.icount.rx++;

			if( !uart_handle_sysrq_char( &sport->port, c, regs ) ) {
				*tty->flip.flag_buf_ptr++ = flag;
				*tty->flip.char_buf_ptr++ = c;
				tty->flip.count++;
			}
		}

		/* check for next characters */

		stat = readl( addr_ser_channel( NS_SER_STAT_A,
						sport->hw_port ) );
		if ( ! (stat & NS_SER_STAT_A_RBC ) )
			break;

		writel( NS_SER_STAT_A_RBC,
			addr_ser_channel( NS_SER_STAT_A, sport->hw_port ) );
		stat = readl( addr_ser_channel( NS_SER_STAT_A,
						sport->hw_port ) );
	}

	tty_flip_buffer_push( tty );

#ifdef SUPPORT_SYSRQ
	sport->port.sysrq = 0;
#endif
}

/***********************************************************************
 * @Function: ns_tx_chars
 * @Return: n/a
 * @Descr: sends as many characters to FIFO as possible.
 ***********************************************************************/

static void ns_ser_tx_chars( ns_ser_port* sport )
{
	struct circ_buf* xmit = &sport->port.info->xmit;

	if( sport->port.x_char ) {
		ns_ser_tx_char( sport, sport->port.x_char );
		sport->port.icount.tx++;
		sport->port.x_char = 0;
		return;
	}

	if( uart_circ_empty( xmit ) || uart_tx_stopped( &sport->port ) ) {
		ns_ser_stop_tx( &sport->port, 0 );
		return;
	}

	do {
		ns_ser_tx_char( sport, xmit->buf[ xmit->tail ] );
		xmit->tail = (xmit->tail + 1) & (UART_XMIT_SIZE - 1);
		sport->port.icount.tx++;
		if( uart_circ_empty( xmit ) )
			break;
	} while(  (readl( addr_ser_channel( NS_SER_STAT_A, sport->hw_port ) ) &
		 NS_SER_STAT_A_THALF ) );

	if( uart_circ_chars_pending( xmit ) < WAKEUP_CHARS )
		uart_write_wakeup( &sport->port );

	if( uart_circ_empty( xmit ) )
		ns_ser_stop_tx( &sport->port, 0 );
}

/***********************************************************************
 * @Function: ns_ser_tx_char
 * @Return: nothing
 * @Descr: sends one character (in interrupt handler
 ***********************************************************************/

inline static void ns_ser_tx_char( ns_ser_port* sport, char c )
{
	ns_ser_wait_for_xmitr( sport->hw_port );
	writeb( c, addr_ser_channel( NS_SER_FIFO, sport->hw_port ) );
}

static void ns_ser_set_bitrate( int cflag, int quot, int hw_port )
{
	unsigned int ctrl_a;
	unsigned int bitrate;

	ctrl_a = NS_SER_CTRL_A_CE;
	/* byte size and parity */
	switch( cflag & CSIZE )
	{
	    case CS5: ctrl_a |= NS_SER_CTRL_A_WLS_5; break;
	    case CS6: ctrl_a |= NS_SER_CTRL_A_WLS_6; break;
	    case CS7: ctrl_a |= NS_SER_CTRL_A_WLS_7; break;
	    case CS8: ctrl_a |= NS_SER_CTRL_A_WLS_8; break;
	}
	if( cflag & CSTOPB )
		ctrl_a |= NS_SER_CTRL_A_STOP;
	if( cflag & PARENB )
		ctrl_a |= NS_SER_CTRL_A_PE;
	if( !(cflag & PARODD) )
		ctrl_a |= NS_SER_CTRL_A_EPS;
	/* for PARODD, no value has to be set */
#if defined(USE_BBUS_CLK)
	bitrate = ( NS_SER_BITRATE_EBIT |
		    NS_SER_BITRATE_CLKMUX_BCLK |
		    NS_SER_BITRATE_TMODE |
		    NS_SER_BITRATE_TCDR_16 |
		    NS_SER_BITRATE_RCDR_16 |
		    ( ( quot - 1 ) &
		      NS_SER_BITRATE_N_MA ) );
#else
	bitrate = ( NS_SER_BITRATE_EBIT |
		    NS_SER_BITRATE_TMODE |
		    NS_SER_BITRATE_TCDR_16 |
		    NS_SER_BITRATE_RCDR_16 |
		    ( ( quot - 1 ) &
		      NS_SER_BITRATE_N_MA ) );
#endif

	writel( (readl( addr_ser_channel( NS_SER_CTRL_A, hw_port ) ) &
		 ~(NS_SER_CTRL_A_WLS_8 |
		   NS_SER_CTRL_A_STOP |
		   NS_SER_CTRL_A_PE |
		   NS_SER_CTRL_A_EPS)) | ctrl_a,
		 addr_ser_channel( NS_SER_CTRL_A, hw_port ) );

	writel( bitrate, addr_ser_channel( NS_SER_BITRATE, hw_port ));
	/* @TODO: when chip supports it, do real calculation */
	writel( NS_SER_RX_CHAR_TIMER_TRUN,
		addr_ser_channel( NS_SER_RX_CHAR_TIMER, hw_port ) );
}

#ifdef CONFIG_SERIAL_NS9750_CONSOLE

struct baud_rates {
	unsigned int rate;
	unsigned int cflag;
};

static struct baud_rates baud_rates[] = {
	{ 921600, B921600 },
	{ 460800, B460800 },
	{ 230400, B230400 },
	{ 115200, B115200 },
	{  57600, B57600  },
	{  38400, B38400  },
	{  19200, B19200  },
	{   9600, B9600   },
	{   4800, B4800   },
	{   2400, B2400   },
	{   1200, B1200   },
	{      0, B38400  }
};

static int __init ns_console_init( void )
{
	register_console( &ns_console );
	return 0;
}

#if 0
/*
 * If the port was already initialised (eg, by a boot loader),
 * try to determine the current setup.
 */
static void __init ns_console_get_options( int* baud, int* parity, int* bits,
					   int hw_port)
{
	unsigned int ctrl_a;
	unsigned int bitrate;

	/* no need to worry about other processes accessing bitrate */
	ctrl_a = readl( addr_ser_channel( NS_SER_CTRL_A, hw_port ) );
	bitrate = readl( addr_ser_channel( NS_SER_BITRATE, hw_port ) );

	if( !(ctrl_a & NS_SER_CTRL_A_CE ) )
		/* not enabled, don't modify defaults */
		return;

	switch( ctrl_a & NS_SER_CTRL_A_WLS_MA ) {
	    case NS_SER_CTRL_A_WLS_5 : *bits = 5; break;
	    case NS_SER_CTRL_A_WLS_6 : *bits = 6; break;
	    case NS_SER_CTRL_A_WLS_7 : *bits = 7; break;
	    case NS_SER_CTRL_A_WLS_8 :
	    default: *bits = 8; break;
	}

	switch( ctrl_a & NS_SER_CTRL_A_PE ) {
	    case NS_SER_CTRL_A_PE : *parity = 'e'; break;
	    case (NS_SER_CTRL_A_PE|NS_SER_CTRL_A_EPS) : *parity = 'o'; break;
	    default: *parity = 'n'; break;
	}

	/* expects a clk divide rate of 16, no check is done yet */
	*baud = BBUS_CLK_FREQ/(((bitrate & NS_SER_BITRATE_N_MA) + 1)*16);
}
#endif

static int __init ns_console_setup( struct console* co, char* options )
{
	ns_ser_port* sport;
	int baud = 115200;
	int bits = 8;
	int parity = 'n';
	int flow = 'n';
	int cflag;
	int ret;
	int i;

	/* only accept supported UARTs, otherwise take first */
	if( co->index == -1 || co->index >= NR_PORTS )
		co->index = 0;

	ns_ser_init_ports( co->index );

	sport = &ns_ser_ports[ co->index ];

	if( options )
		uart_parse_options( options, &baud, &parity, &bits, &flow );

	/* create cflag as co->cflag may not be set if no options */
	cflag = 0;
	switch( bits )
	{
	    case 5: cflag |= CS5; break;
	    case 6: cflag |= CS6; break;
	    case 7: cflag |= CS7; break;
	    case 8:
	    default: cflag |= CS8; break;
	}
	switch( parity ) {
	    case 'o':
	    case 'O':
		cflag |= PARODD;
		/*fall through*/
	    case 'e':
	    case 'E':
		cflag |= PARENB;
		break;
	}

	set_gpio_cfg( ns_ser_gpios[ co->index ][ GPIO_TX ], NS_GPIO_CFG_FUNC_0 );

	/* no need to worry about other processes accessing bitrate */
#if defined(USE_BBUS_CLK)
	ns_ser_set_bitrate( cflag, BBUS_CLK_FREQ / ( baud * 16 ), co->index );
#else
	ns_ser_set_bitrate( cflag, (XTAL_FREQ / 2) / (baud * 16), co->index );
#endif
	ret = uart_set_options( &sport->port, co, baud, parity, bits, flow );

	/* remember baudrate for termios. this will fail if multiple consoles
	 * are used for this driver with different baudrates, only the last one
	 * will be used for all */
	for( i = 0; baud_rates[i].rate; i++ )
		if( baud_rates[i].rate <= baud )
			break;

	ns_console_cbaud = baud_rates[i].cflag;

	/* for default consoles in kernel commandline, co is not initialized */
	return ret;
}

/*
 * Interrupts are disabled on entering
 */
static void ns_console_write( struct console* co, const char* s,
			      unsigned int count )
{
	char c = *s;
	unsigned int ctrl_old;
	unsigned int ctrl;

	/* save state and disable interrupts */
	/* we keep the baudrate settings of /dev/ttyAM0 */
	ctrl_old = readl( addr_ser_channel( NS_SER_CTRL_A, co->index ) );
	/* may be disabled, so renable it  */
	ctrl = (ctrl_old | NS_SER_CTRL_A_CE) & ~NS_SER_CTRL_A_TIC_MA &
		~NS_SER_CTRL_A_RIC_MA & ~NS_SER_CTRL_A_RIE_MA;
	writel( ctrl, addr_ser_channel( NS_SER_CTRL_A, co->index ) );

	while( count ) {
		ns_ser_wait_for_xmitr( co->index );
		writeb( c, addr_ser_channel( NS_SER_FIFO, co->index ) );
		if( c == '\n' )
			c = '\r';
		else {
			/* next character */
			count--;
			s++;
			c = *s;
		}
	}

	if( !(ctrl_old & NS_SER_CTRL_A_CE ) )
		/* send whole message  */
		ns_ser_wait_for_fifo_empty( co->index );

	/* enable interrupts */
	writel( ctrl_old, addr_ser_channel( NS_SER_CTRL_A, co->index ) );
}

console_initcall( ns_console_init );
#endif /* CONFIG_SERIAL_NS9750_CONSOLE */

module_init( ns_ser_init );
module_exit( ns_ser_exit );

MODULE_AUTHOR( "Markus Pietrek, FS Forth-Systeme GmbH" );
MODULE_DESCRIPTION( "NS9xx0 serial driver $Revision: 1.6 $" );
MODULE_LICENSE( "GPL" );
